import os


class FileMapper:
    """
    Class to represent a set of files that exist in the same directory as one object.

    Args:
        file_list: the list of files to represent. If dirpath not providedm these should be full file paths.
        dirpath: The directory path where the files exist. Default is None.

    If dirpath is not provided it will be deduced from the file paths provided in file_list.

    Attributes:
        file_list: list of file names of the files represented.
        file_paths: list of full file paths of the files represented.
        dirpath: The directory path where the files exist. Either deduced or provided.


    """

    def __init__(self, file_list, dirpath=None):
        self.dirpath = dirpath
        self.file_list = file_list

        self._resolve()

    def _resolve(self):
        if not self.dirpath:
            first_dir = os.path.dirname(self.file_list[0])
            if os.path.dirname(os.path.commonprefix(self.file_list)) == first_dir:
                self.dirpath = first_dir
            else:
                raise Exception(
                    "File inputs are not from the same directory so cannot be resolved."
                )

        self.file_list = [os.path.basename(fpath) for fpath in self.file_list]
        self.file_paths = [
            os.path.join(self.dirpath, fname) for fname in self.file_list
        ]
        # check all files exist on file system
        if not all(os.path.isfile(file) for file in self.file_paths):
            raise FileNotFoundError("Some files could not be found.")


def is_file_list(coll):
    """
    Checks whether a collection is a list of files.

    :param coll (list): collection to check.
    :return: True if collection is a list of files, else returns False.
    """
    # check if collection is a list of files
    if not isinstance(coll, list):
        raise Exception(f"Expected collection as a list, have received {type(coll)}")

    if coll[0].startswith("/") or coll[0].startswith("http"):
        return True

    return False



import os
from math import isclose

import numpy as np
import pytest
import xarray as xr

from clisops.utils.dataset_utils import get_coord_by_type


class TestRoll:

    @staticmethod
    def calculate_offset(x, lon):
        # get resolution of data
        res = lon.values[1] - lon.values[0]

        # work out how many to roll by to roll data by 1
        index = 1 / res

        # calculate the corresponding offset needed to change data by x
        offset = int(x * index)

        return offset

    def test_roll_lon_minus_180(self, mini_esgf_data):
        ds = xr.open_dataset(mini_esgf_data["CMIP6_RLDS_ONE_TIME_STEP"])
        # gets longitude by the correct name as used in the dataset
        lon = ds[get_coord_by_type(ds, "longitude")]

        # check longitude is 0 to 360 initially
        assert isclose(lon.values.min(), 0, abs_tol=10**2)
        assert isclose(lon.values.max(), 360, abs_tol=10**2)

        # roll longitude by -180
        ds = ds.roll(shifts={f"{lon.name}": -180}, roll_coords=True)

        # doesn't roll as far as we want as it hasn't taken the resolution of the grid into account
        assert ds.lon.values[0] == 90.0
        assert ds.lon.values[-1] == 87.5

        # min and max of the data are still the same
        assert ds.lon.values.min() == 0
        assert ds.lon.values.max() == 357.5

    def test_roll_lon_minus_180_use_res(self, mini_esgf_data):
        ds = xr.open_dataset(mini_esgf_data["CMIP6_RLDS_ONE_TIME_STEP"])
        # gets longitude by the correct name as used in the dataset
        lon = ds[get_coord_by_type(ds, "longitude")]

        # work out how much to roll by
        offset = self.calculate_offset(-180, lon)

        # roll longitude by calculated offset
        ds_roll = ds.roll(shifts={f"{lon.name}": offset}, roll_coords=True)

        # longitude data array is rolled to [180.0..275.0..357.5,0..75..177.5]
        assert ds_roll.lon.values[0] == 180.0
        assert ds_roll.lon.values[-1] == 177.5

        # min and max still the same
        assert ds_roll.lon.values.min() == 0
        assert ds_roll.lon.values.max() == 357.5

        # rlds values are not equal - they have been rolled
        np.testing.assert_raises(
            AssertionError,
            np.testing.assert_array_equal,
            ds_roll.rlds.values,
            ds.rlds.values,
        )

    def test_roll_lon_plus_180(self, mini_esgf_data):
        ds = xr.open_dataset(mini_esgf_data["CMIP6_RLDS_ONE_TIME_STEP"])
        # gets longitude by the correct name as used in the dataset
        lon = ds[get_coord_by_type(ds, "longitude")]

        ds = ds.roll(shifts={f"{lon.name}": 180}, roll_coords=True)

        # doesn't roll as far as we want as it hasn't taken the resolution of the grid into account
        assert ds.lon.values[0] == 270.0
        assert ds.lon.values[-1] == 267.5

        assert ds.lon.values.min() == 0
        assert ds.lon.values.max() == 357.5

    def test_roll_lon_plus_180_use_res(self, mini_esgf_data):
        ds = xr.open_dataset(mini_esgf_data["CMIP6_RLDS_ONE_TIME_STEP"])
        # gets longitude by the correct name as used in the dataset
        lon = ds[get_coord_by_type(ds, "longitude")]

        # work out how much to roll by
        offset = self.calculate_offset(180, lon)

        ds = ds.roll(shifts={f"{lon.name}": offset}, roll_coords=True)

        assert ds.lon.values[0] == 180.0
        assert ds.lon.values[-1] == 177.5

        assert ds.lon.values.min() == 0
        assert ds.lon.values.max() == 357.5

    def test_plus_minus_180_equal(self, mini_esgf_data):
        # check that rolling +180 and -180 gives the same result - when taking the resolution into account
        ds = xr.open_dataset(mini_esgf_data["CMIP6_RLDS_ONE_TIME_STEP"])
        # gets longitude by the correct name as used in the dataset
        lon = ds[get_coord_by_type(ds, "longitude")]

        # work out how much to roll by
        offset_minus = self.calculate_offset(-180, lon)
        offset_plus = self.calculate_offset(180, lon)

        ds_minus = ds.roll(shifts={f"{lon.name}": offset_minus}, roll_coords=True)
        ds_plus = ds.roll(shifts={f"{lon.name}": offset_plus}, roll_coords=True)

        # values of rlds are equal - rolling by -180 and 180 (taking res into account) is the same
        np.testing.assert_allclose(ds_minus.rlds.values, ds_plus.rlds.values)

    def test_roll_compare_roll_coords(self, mini_esgf_data):
        ds = xr.open_dataset(mini_esgf_data["CMIP6_RLDS_ONE_TIME_STEP"])
        # gets longitude by the correct name as used in the dataset
        lon = ds[get_coord_by_type(ds, "longitude")]

        # work out how much to roll by
        offset = self.calculate_offset(180, lon)

        ds_roll_coords = ds.roll(shifts={f"{lon.name}": offset}, roll_coords=True)
        ds_not_roll_coords = ds.roll(shifts={f"{lon.name}": offset}, roll_coords=False)

        # check rlds values the same with/without rolling coords
        np.testing.assert_array_equal(
            ds_roll_coords.rlds.values,
            ds_not_roll_coords.rlds.values,
        )

        # check lat doesn't change with/without rolling coords
        np.testing.assert_array_equal(
            ds_roll_coords.lat.values,
            ds_not_roll_coords.lat.values,
        )

        # check time doesn't change with/without rolling coords
        np.testing.assert_array_equal(
            ds_roll_coords.time.values,
            ds_not_roll_coords.time.values,
        )

        # check lon changes with/without rolling coords
        np.testing.assert_raises(
            AssertionError,
            np.testing.assert_array_equal,
            ds_roll_coords.lon.values,
            ds_not_roll_coords.lon.values,
        )


def test_irregular_grid_dataset(mini_esgf_data):
    ds = xr.open_dataset(mini_esgf_data["CMIP6_SIMASS_DEGEN"])
    lon = get_coord_by_type(ds, "longitude", ignore_aux_coords=False)

    assert "lon" not in ds.dims

    with pytest.raises(ValueError) as exc:
        ds.roll(shifts={f"{lon}": 180}, roll_coords=False)
    assert str(exc.value) in [
        "dimensions ['longitude'] do not exist",
        "Dimensions ['longitude'] not found in data dimensions ('i', 'j', 'time', 'bnds', 'vertices')",
    ]


def test_3d_grid_dataset(mini_esgf_data):
    ds = xr.open_mfdataset(mini_esgf_data["CMIP6_TA"])
    lon = get_coord_by_type(ds, "longitude", ignore_aux_coords=False)

    assert "lon" in ds.dims

    offset = 180

    ds_roll_coords = ds.roll(shifts={f"{lon}": offset}, roll_coords=True)
    ds_not_roll_coords = ds.roll(shifts={f"{lon}": offset}, roll_coords=False)

    # check plev doesn't change with/without rolling coords
    np.testing.assert_array_equal(
        ds_roll_coords.plev.values,
        ds_not_roll_coords.plev.values,
    )



import warnings
from datetime import datetime as dt
from pathlib import Path
from typing import Optional, Union

import xarray as xr
from loguru import logger

from clisops.core import Grid, Weights
from clisops.core import regrid as core_regrid
from clisops.exceptions import InvalidParameterValue
from clisops.ops.base_operation import Operation
from clisops.utils.file_namers import get_file_namer

__all__ = [
    "regrid",
]

supported_regridding_methods = ["conservative", "patch", "nearest_s2d", "bilinear"]


class Regrid(Operation):
    """Class for regridding operation, extends clisops.ops.base_operation.Operation."""

    @staticmethod
    def _get_grid_in(
        grid_desc: Union[xr.Dataset, xr.DataArray],
        compute_bounds: bool,
    ):
        """
        Create clisops.core.regrid.Grid object as input grid of the regridding operation.

        Return the Grid object.
        """
        if isinstance(grid_desc, (xr.Dataset, xr.DataArray)):
            return Grid(ds=grid_desc, compute_bounds=compute_bounds)
        raise InvalidParameterValue(
            "An xarray.Dataset or xarray.DataArray has to be provided as input for the source grid."
        )

    def _get_grid_out(
        self,
        grid_desc: Union[xr.Dataset, xr.DataArray, int, float, tuple, str],
        compute_bounds: bool,
        mask: Optional[str] = None,
    ) -> Grid:
        """Create clisops.core.regrid.Grid object as target grid of the regridding operation.

        Returns
        -------
        Grid
        """
        if isinstance(grid_desc, str):
            if grid_desc in ["auto", "adaptive"]:
                return Grid(
                    ds=self.ds,
                    grid_id=grid_desc,
                    compute_bounds=compute_bounds,
                    mask=mask,
                )
            else:
                return Grid(grid_id=grid_desc, compute_bounds=compute_bounds, mask=mask)
        elif isinstance(grid_desc, (float, int, tuple)):
            return Grid(
                grid_instructor=grid_desc, compute_bounds=compute_bounds, mask=mask
            )
        elif isinstance(grid_desc, (xr.Dataset, xr.DataArray)):
            return Grid(ds=grid_desc, compute_bounds=compute_bounds, mask=mask)
        else:
            # clisops.core.regrid.Grid will raise the exception
            return Grid()

    @staticmethod
    def _get_weights(grid_in: Grid, grid_out: Grid, method: str):
        """Generate the remapping weights using clisops.core.regrid.Weights.

        Returns
        -------
        Weights
            An instance of the Weights object.
        """
        return Weights(grid_in=grid_in, grid_out=grid_out, method=method)

    def _resolve_params(self, **params) -> None:
        """Generate a dictionary of regrid parameters."""
        # all regrid specific parameters should be passed in via **params
        # this is where we resolve them and set self.params as a dict or as separate attributes
        # this would be where you make use of your other methods/ attributes e.g.
        # get_grid_in(), get_grid_out() and get_weights() to generate the regridder

        adaptive_masking_threshold = params.get("adaptive_masking_threshold", None)
        grid = params.get("grid", None)
        method = params.get("method", None)
        keep_attrs = params.get("keep_attrs", None)
        mask = params.get("mask", None)

        if mask not in ["land", "ocean", False, None]:
            raise ValueError(
                f"mask must be one of 'land', 'ocean' or None, not '{mask}'."
            )

        if method not in supported_regridding_methods:
            raise Exception(
                "The selected regridding method is not supported. "
                "Please choose one of %s." % ", ".join(supported_regridding_methods)
            )

        logger.debug(
            f"Input parameters: method: {method}, grid: {grid}, adaptive_masking: {adaptive_masking_threshold}, mask: {mask}, keep_attrs: {keep_attrs}"
        )

        # Compute bounds only when required
        compute_bounds = "conservative" in method

        # Create and check source and target grids
        grid_in = self._get_grid_in(self.ds, compute_bounds)
        grid_out = self._get_grid_out(grid, compute_bounds, mask=mask)

        if grid_in.hash == grid_out.hash:
            weights = None
            regridder = None
            weights_filename = None
        else:
            # Compute the remapping weights
            t_start = dt.now()
            weights = self._get_weights(
                grid_in=grid_in, grid_out=grid_out, method=method
            )
            regridder = weights.regridder
            weights_filename = regridder.filename
            t_end = dt.now()
            logger.info(
                f"Computed/Retrieved weights in {(t_end - t_start).total_seconds()} seconds."
            )

        # Define params dict
        self.params = {
            "orig_ds": self.ds,
            "grid_in": grid_in,
            "grid_out": grid_out,
            "method": method,
            "regridder": regridder,
            "weights": weights,
            "adaptive_masking_threshold": adaptive_masking_threshold,
            "keep_attrs": keep_attrs,
        }

        # Input grid / Dataset
        self.ds = self.params.get("grid_in").ds

        # There is no __str__() method for the Regridder object, so I used its filename attribute,
        # which specifies a default filename (does not correspond with the filename we would give the weight file).
        # todo: Better option might be to have the Weights class extend the Regridder class or to define
        #  a __str__() method for the Weights class.
        logger.debug(
            "Resolved parameters: grid_in: {}, grid_out: {}, regridder: {}".format(
                self.params.get("grid_in").__str__(),
                self.params.get("grid_out").__str__(),
                weights_filename,
            )
        )

    def _get_file_namer(self) -> object:
        """Return the appropriate file namer object."""
        # "extra" is what will go at the end of the file name before .nc
        extra = "_regrid-{}-{}".format(
            self.params.get("method"), self.params.get("grid_out").__str__()
        )

        namer = get_file_namer(self._file_namer)(extra=extra)

        return namer

    def _calculate(self):
        """
        Process the regridding request, calls clisops.core.regrid.regrid().

        Returns the resulting xarray.Dataset.
        """

        # Pass through the input dataset if grid_in and grid_out are equal
        if self.params.get("grid_in").hash == self.params.get("grid_out").hash:
            warnings.warn(
                "The selected source and target grids are the same. "
                "No regridding operation required."
            )
            return self.params.get("orig_ds")

        # the result is saved by the process() method on the base class
        regridded_ds = core_regrid(
            self.params.get("grid_in", None),
            self.params.get("grid_out", None),
            self.params.get("weights", None),
            self.params.get("adaptive_masking_threshold", None),
            self.params.get("keep_attrs", None),
        )

        return regridded_ds


def regrid(
    ds: Union[xr.Dataset, str, Path],
    *,
    method: Optional[str] = "nearest_s2d",
    adaptive_masking_threshold: Optional[Union[int, float]] = 0.5,
    grid: Optional[
        Union[xr.Dataset, xr.DataArray, int, float, tuple, str]
    ] = "adaptive",
    mask: Optional[str] = None,
    output_dir: Optional[Union[str, Path]] = None,
    output_type: Optional[str] = "netcdf",
    split_method: Optional[str] = "time:auto",
    file_namer: Optional[str] = "standard",
    keep_attrs: Optional[Union[bool, str]] = True,
) -> list[Union[xr.Dataset, str]]:
    """Regrid specified input file or xarray object.

    Parameters
    ----------
    ds : Union[xr.Dataset, str]
    method : {"nearest_s2d", "conservative", "patch", "bilinear"}
    adaptive_masking_threshold : Optional[Union[int, float]]
    grid : Union[xr.Dataset, xr.DataArray, int, float, tuple, str]
    mask: {"ocean", "land"} = None
    output_dir : Optional[Union[str, Path]] = None
    output_type : {"netcdf", "nc", "zarr", "xarray"}
    split_method : {"time:auto"}
    file_namer : {"standard", "simple"}
    keep_attrs : {True, False, "target"}

    Returns
    -------
    List[Union[xr.Dataset, str]]
        A list of the regridded outputs in the format selected; str corresponds to file paths if the
        output format selected is a file.

    Examples
    --------
    | ds: xarray Dataset or "cmip5.output1.MOHC.HadGEM2-ES.rcp85.mon.atmos.Amon.r1i1p1.latest.tas"
    | method: "nearest_s2d"
    | adaptive_masking_threshold:
    | grid: "1deg"
    | mask: "land"
    | output_dir: "/cache/wps/procs/req0111"
    | output_type: "netcdf"
    | split_method: "time:auto"
    | file_namer: "standard"
    | keep_attrs: True
    """
    op = Regrid(**locals())
    return op.process()



import inspect
import os
import warnings
from typing import Optional

import cf_xarray as cfxr  # noqa
import cftime
import dask.array as da
import fsspec
import numpy as np
import xarray as xr

from clisops.exceptions import InvalidParameterValue
from clisops.project_utils import dset_to_filepaths
from clisops.utils.time_utils import str_to_AnyCalendarDateTime

known_coord_types = ["time", "level", "latitude", "longitude", "realization"]

KERCHUNK_EXTS = [".json", ".zst", ".zstd"]


def get_coord_by_type(
    ds, coord_type, ignore_aux_coords=True, return_further_matches=False
):
    """
    Returns the name of the coordinate that matches the given type.

    Parameters
    ----------
    ds : xarray.Dataset or xarray.DataArray
        Dataset/DataArray to search for coordinate.
    coord_type : str
        Type of coordinate, e.g. 'time', 'level', 'latitude', 'longitude', 'realization'.
    ignore_aux_coords : bool
        Whether to ignore auxiliary coordinates.
    return_further_matches : bool
        Whether to return further matches.

    Returns
    -------
    str
        Name of the coordinate that matches the given type.
    str, list of str
        If return_further_matches is True, apart from the matching coordinate,
        a list with further potential matches is returned.

    Raises
    ------
    ValueError
        If the coordinate type is not known.
    """
    # List for all potential matches
    coords = list()

    # If coord_type is not in known_coord_types then raise an error
    if coord_type not in known_coord_types:
        raise ValueError(f"Coordinate type not known: {coord_type}")

    # Get main variable ... if possible
    try:
        main_var = get_main_variable(ds)
    except ValueError:
        warnings.warn(f"No main variable found for dataset '{ds}'.")
        main_var = None

    # Loop through all (potential) coordinates to find all possible matches
    if isinstance(ds, xr.DataArray):
        coord_vars = list(ds.coords)
    elif isinstance(ds, xr.Dataset):
        # Not all coordinate variables are always classified as such
        coord_vars = list(ds.coords) + list(ds.data_vars)
        # make sure we skip the main variable!
        if main_var is not None:
            coord_vars.remove(main_var)
    else:
        raise TypeError("Not an xarray.Dataset or xarray.DataArray.")
    for coord_id in coord_vars:
        # If ignore_aux_coords is True then: ignore coords that are not dimensions
        if ignore_aux_coords and coord_id not in ds.dims:
            continue

        coord = ds[coord_id]

        if get_coord_type(coord) == coord_type:
            coords.append(coord_id)

    # Return None if no match
    if len(coords) == 0:
        warnings.warn(f"No coordinate variable found for type '{coord_type}'.")
        return None
    elif len(coords) == 1:
        if return_further_matches:
            return coords[0], []
        else:
            return coords[0]
    # If more than one match is found, a selection has to be made
    else:
        warnings.warn(
            f"More than one coordinate variable found for type '{coord_type}'. Selecting the best fit."
        )
        # Sort in terms of number of dimensions
        coords = sorted(coords, key=lambda x: len(ds[x].dims), reverse=True)

        if main_var is not None:
            # Get dimensions and singleton coords of main variable
            main_var_dims = list(ds[main_var].dims)

            # Select coordinate with most dims (matching with main variable dims)
            for coord_id in coords:
                if coord_id in ds.coords:
                    if all([dim in main_var_dims for dim in ds.coords[coord_id].dims]):
                        if return_further_matches:
                            return coord_id, [x for x in coords if x != coord_id]
                        else:
                            return coord_id
        # If the decision making fails, pass the first match
        if return_further_matches:
            return coords[0], coords[1:]
        else:
            return coords[0]


# from dachar
def get_coord_by_attr(ds, attr, value):
    """
    Returns a coordinate based on a known attribute of a coordinate.

    :param ds: Xarray Dataset or DataArray
    :param attr: (str) Name of attribute to look for.
    :param value: Expected value of attribute you are looking for.
    :return: Coordinate of xarray dataset if found.
    """
    coords = ds.coords

    for coord in coords.values():
        if coord.attrs.get(attr, None) == value:
            return coord

    return None


def is_latitude(coord):
    """
    Determines if a coordinate is latitude.

    :param coord: coordinate of xarray dataset e.g. coord = ds.coords[coord_id]
    :return: (bool) True if the coordinate is latitude.
    """

    if (
        "latitude" in coord.cf.coordinates
        and coord.name in coord.cf.coordinates["latitude"]
    ):
        return True

    if (
        "latitude" in coord.cf.standard_names
        and coord.name in coord.cf.standard_names["latitude"]
    ):
        return True

    if hasattr(coord, "standard_name") and coord.standard_name == "latitude":
        return True

    if hasattr(coord, "long_name") and coord.long_name == "latitude":
        return True

    return False


def is_longitude(coord):
    """
    Determines if a coordinate is longitude.

    :param coord: coordinate of xarray dataset e.g. coord = ds.coords[coord_id]
    :return: (bool) True if the coordinate is longitude.
    """
    if (
        "longitude" in coord.cf.coordinates
        and coord.name in coord.cf.coordinates["longitude"]
    ):
        return True

    if (
        "longitude" in coord.cf.standard_names
        and coord.name in coord.cf.standard_names["longitude"]
    ):
        return True

    if hasattr(coord, "standard_name") and coord.standard_name == "longitude":
        return True

    if hasattr(coord, "long_name") and coord.long_name == "longitude":
        return True

    return False


def is_level(coord):
    """
    Determines if a coordinate is level.

    :param coord: coordinate of xarray dataset e.g. coord = ds.coords[coord_id]
    :return: (bool) True if the coordinate is level.
    """
    if (
        "vertical" in coord.cf.coordinates
        and coord.name in coord.cf.coordinates["vertical"]
    ):
        return True

    if hasattr(coord, "positive"):
        if coord.attrs.get("positive", None) == "up" or "down":
            return True

    if hasattr(coord, "axis"):
        if coord.attrs.get("axis", None) == "Z":
            return True

    return False


def _is_time(coord):
    """
    Check if a coordinate uses cftime datetime objects.

    Handles Dask-backed arrays for lazy evaluation.
    """
    if coord.size == 0:
        return False  # Empty array

    if isinstance(coord.dtype.type(), cftime.datetime):
        return True

    # Safely get first element without loading entire array
    first_value = coord.isel({dim: 0 for dim in coord.dims}).values

    # Compute only if it's a Dask array
    if isinstance(first_value, da.Array):
        first_value = first_value.compute()

    return isinstance(first_value.item(0), cftime.datetime)


def is_time(coord):
    """
    Determines if a coordinate is time.

    :param coord: coordinate of xarray dataset e.g. coord = ds.coords[coord_id]
    :return: (bool) True if the coordinate is time.
    """
    if False and coord.ndim >= 2:
        # skip variables with more than two dimensions: lat_bnds, lon_bnds, time_bnds, t, ...
        return False

    if "time" in coord.cf.coordinates and coord.name in coord.cf.coordinates["time"]:
        return True

    if (
        "time" in coord.cf.standard_names
        and coord.name in coord.cf.standard_names["time"]
    ):
        return True

    if np.issubdtype(coord.dtype, np.datetime64):
        return True

    if hasattr(coord, "axis"):
        if coord.axis == "T":
            return True

    return _is_time(coord)


def is_realization(coord):
    """
    Determines if a coordinate is realization.

    :param coord: coordinate of xarray dataset e.g. coord = ds.coords[coord_id]
    :return: (bool) True if the coordinate is longitude.
    """
    if (
        "realization" in coord.cf.standard_names
        and coord.name in coord.cf.standard_names["realization"]
    ):
        return True

    if coord.attrs.get("standard_name", None) == "realization":
        return True

    return False


def get_coord_type(coord):
    """
    Gets the coordinate type.

    :param coord: coordinate of xarray dataset e.g. coord = ds.coords[coord_id]
    :return: The type of coordinate as a string. Either longitude, latitude, time, level or None
    """

    if is_longitude(coord):
        return "longitude"
    elif is_latitude(coord):
        return "latitude"
    elif is_level(coord):
        return "level"
    elif is_time(coord):
        return "time"
    elif is_realization(coord):
        return "realization"

    return None


def get_main_variable(ds, exclude_common_coords=True):
    """
    Finds the main variable of an xarray Dataset

    :param ds: xarray Dataset
    :param exclude_common_coords: (bool) If True then common coordinates are excluded from the search for the
                                main variable. common coordinates are time, level, latitude, longitude and bounds.
                                Default is True.
    :return: (str) The main variable of the dataset e.g. 'tas'
    """
    if isinstance(ds, xr.Dataset):
        variables = list(ds.variables.items())
        data_dims = [data.dims for var_id, data in variables]
    else:
        variables = []
        data_dims = []
    flat_dims = [dim for sublist in data_dims for dim in sublist]

    results = {}
    common_coords = [
        "bnd",
        "bound",
        "lat",
        "lon",
        "time",
        "level",
        "realization_index",
        "realization",
    ]

    for var_id, data in variables:
        if var_id in flat_dims:
            continue
        if exclude_common_coords is True and any(
            coord in var_id for coord in common_coords
        ):
            continue
        else:
            results.update({var_id: len(ds[var_id].shape)})
    result = max(results, key=results.get)

    if result is None:
        raise Exception("Could not determine main variable")
    else:
        return result


def open_xr_dataset(dset, **kwargs):
    """
    Opens an xarray dataset from a dataset input.

    :param dset: (str or Path) A dataset identifier, directory path, or file path ending in ``*.nc``.
    :param kwargs: Any additional keyword arguments for opening the dataset.
                   ``decode_times=xr.coders.CFDatetimeCoder(use_cftime=True)`` and ``decode_timedelta=False`` are used by default,
                   along with ``combine="by_coords"`` for ``open_mfdataset`` only.

    Notes:
        - Any list will be interpreted as a list of files.
    """
    # Set up dictionaries of arguments to send to all `xr.open_*dataset()` calls
    single_file_kwargs = _get_kwargs_for_opener("single", **kwargs)
    multi_file_kwargs = _get_kwargs_for_opener("multi", **kwargs)

    # Assume that a JSON or ZST/ZSTD file is kerchunk
    if type(dset) not in (list, tuple):
        # Assume that a JSON or ZST/ZSTD file is kerchunk
        if is_kerchunk_file(dset):
            return _open_as_kerchunk(dset, **single_file_kwargs)

        else:
            # Force the value of dset to be a list if not a list or tuple
            # use force=True to allow all file paths to pass through DatasetMapper
            dset = dset_to_filepaths(dset, force=True)

    # If an empty sequence, then raise an Exception
    if len(dset) == 0:
        raise Exception("No files found to open with xarray.")

    # if a list we want a multi-file dataset
    if len(dset) > 1:
        ds = xr.open_mfdataset(dset, **multi_file_kwargs)
        # Ensure that time units are retained
        _patch_time_encoding(ds, dset, **single_file_kwargs)
        return ds

    # if there is only one file we only need to call open_dataset
    else:
        return xr.open_dataset(dset[0], **single_file_kwargs)


def _get_kwargs_for_opener(otype, **kwargs):
    """
    Returns a dictionary of keyword args for sending to either `xr.open_dataset()`
    of `xr.open_mfdataset()`, based on whether otype="single" or "multi".
    The provided `kwargs` dictionary is used to extend/override the default
    values.

    :param otype: (Str) type of opener (either "single" or "multi")
    :param kwargs: Any further keyword arguments to include when opening the dataset.
    """
    args = {
        "decode_times": xr.coders.CFDatetimeCoder(use_cftime=True),
        "decode_timedelta": False,
    }

    if otype.lower().startswith("multi"):
        args["combine"] = "by_coords"

    args.update(kwargs)

    # If single file opener, then remove any multifile args that would raise an
    # exception when called
    if otype.lower() == "single":
        [
            args.pop(arg)
            for arg in list(args)
            if arg not in inspect.getfullargspec(xr.open_dataset).kwonlyargs
        ]

    return args


def is_kerchunk_file(dset):
    """
    Returns a boolean based on reading the file extension.
    """
    if not isinstance(dset, str):
        return False

    return os.path.splitext(dset)[-1] in KERCHUNK_EXTS


def _open_as_kerchunk(dset, **kwargs):
    """
    Open the dataset `dset` as a Kerchunk file. Return an Xarray Dataset.
    """
    compression = (
        "zstd"
        if dset.split(".")[-1].startswith("zst")
        else kwargs.get("compression", None)
    )
    remote_options = kwargs.get("remote_options", {})
    remote_protocol = kwargs.get("remote_protocol", None)

    mapper = fsspec.get_mapper(
        "reference://",
        fo=dset,
        target_options={"compression": compression},
        remote_options=remote_options,
        remote_protocol=remote_protocol,
    )

    # Create a copy of kwargs and remove mapper-specific values
    kw = kwargs.copy()
    for key in ("compression", "remote_options", "remote_protocol"):
        if key in kw:
            del kw[key]

    return xr.open_zarr(mapper, consolidated=False, **kw)


def _patch_time_encoding(ds, file_list, **kwargs):
    """
    NOTE: Hopefully this will be fixed in Xarray at some point. The problem is that if
          time is present, the multi-file dataset has an empty `encoding` dictionary.

    Reads the first file in `file_list` to read in the time units attribute. It then
    saves that attribute in `ds.time.encoding["units"]`.

    :param ds: xarray.Dataset
    :file_list: list of file paths
    """
    # Check that first file exists, if not return
    f1 = sorted(file_list)[0]

    if not os.path.isfile(f1):
        return

    # If time is present and the multi-file dataset has an empty `encoding` dictionary.
    # Open the first file to get the time units and add into encoding dictionary.
    if hasattr(ds, "time") and not ds.time.encoding.get("units"):
        ds1 = xr.open_dataset(f1, **kwargs)
        ds.time.encoding["units"] = ds1.time.encoding.get("units", "")


def convert_coord_to_axis(coord):
    """
    Converts coordinate type to its single character axis identifier (tzyx).

    :param coord: (str) The coordinate to convert.
    :return: (str) The single character axis identifier of the coordinate (tzyx).
    """

    axis_dict = {
        "time": "t",
        "longitude": "x",
        "latitude": "y",
        "level": "z",
        "realization": "r",
    }
    return axis_dict.get(coord, None)


def determine_lon_lat_range(ds, lon, lat, lon_bnds=None, lat_bnds=None, apply_fix=True):
    """Determine the min/max lon/lat values of the dataset (and potentially apply fix for unmasked missing values).

    Parameters
    ----------
    ds : xarray.Dataset
        Input dataset.
    lon : str
        Name of longitude coordinate.
    lat : str
        Name of latitude coordinate.
    lon_bnds : str or None, optional
        Name of longitude bounds coordinate. The default is None.
    lat_bnds : str or None, optional
        Name of latitude bounds coordinate. The default is None.
    apply_fix : bool, optional
        Whether to apply fix for unmasked missing values. The default is True.

    Returns
    -------
    xmin : float
        Minimum longitude value.
    xmax : float
        Maximum longitude value.
    ymin : float
        Minimum latitude value.
    ymax : float
        Maximum latitude value.
    """
    # Determine min/max lon/lat values
    xmin = ds[lon].min().item()
    xmax = ds[lon].max().item()
    ymin = ds[lat].min().item()
    ymax = ds[lat].max().item()

    # Potentially apply fix for unmasked missing values
    if apply_fix:
        if fix_unmasked_missing_values_lon_lat(
            ds, lon, lat, lon_bnds, lat_bnds, [xmin, xmax], [ymin, ymax]
        ):
            xmin = ds[lon].min().item()
            xmax = ds[lon].max().item()
            ymin = ds[lat].min().item()
            ymax = ds[lat].max().item()

    return xmin, xmax, ymin, ymax


def fix_unmasked_missing_values_lon_lat(
    ds, lon, lat, lon_bnds, lat_bnds, xminmax, yminmax
):
    """Fix for unmasked missing values in longitude and latitude coordinates and their bounds

    Parameters
    ----------
    ds : xarray.Dataset
        Input dataset.
    lon : str
        Name of longitude coordinate.
    lat : str
        Name of latitude coordinate.
    lon_bnds : str or None
        Name of longitude bounds coordinate.
    lat_bnds : str or None
        Name of latitude bounds coordinate.
    xminmax : list
        List of minimum and maximum longitude values.
    yminmax : list
        List of minimum and maximum latitude values.

    Returns
    -------
    fix : bool
        Whether the fix on ds[lon] and ds[lat] (and if specified ds[lon_bnds] and ds[lat_bnds]) was applied or not.
    """
    fix = False
    minval = -999
    maxval = 999

    # Potentially fix unmasked missing values in longitude/latitude arrays
    if any([xymin <= minval for xymin in xminmax + yminmax]) or any(
        [xymax >= maxval for xymax in xminmax + yminmax]
    ):

        # Identify potential missing values by detecting outliers
        mask_y = (ds[lat] <= minval) | (ds[lat] >= maxval)
        possible_missing_values_y = ds[lat].where(mask_y)
        mask_x = (ds[lon] <= minval) | (ds[lon] >= maxval)
        possible_missing_values_x = ds[lon].where(mask_x)

        # TBD - potential TODO - Explicitly check the vertices as well for possible missing values
        #                        and not apply the mask from lat / lon.
        #                      - Check if the fields already contain nans (and if they are consistent
        #                        between lat and lon).

        # Find out if the outlier values are unique
        possible_missing_values_x_min = possible_missing_values_x.min().item()
        possible_missing_values_x_max = possible_missing_values_x.max().item()
        possible_missing_values_y_min = possible_missing_values_y.min().item()
        possible_missing_values_y_max = possible_missing_values_y.max().item()

        possible_missing_values = [
            val
            for val in [
                possible_missing_values_x_min,
                possible_missing_values_x_max,
                possible_missing_values_y_min,
                possible_missing_values_y_max,
            ]
            if not np.isnan(val)
        ]

        # Compare the masks for lat / lon and abort the fix if they differ
        if (
            ds[lat].dims != ds[lon].dims
            and len(ds[lon].dims) == 1
            and len(ds[lat].dims) == 1
        ):
            # Abort fix for regular lat-lon grids (1D coordinate variables should not include missing values
            #  - for some of the operations the outliers will cause an exception later on)
            warnings.warn(
                f"Extreme value(s) (potentially unmasked missing_values) found in {lon} and {lat} arrays: "
                f"{set(possible_missing_values)}. A fix is not possible for regular latitude-longitude grids."
            )
            return fix
        elif not (mask_x == mask_y).all().item():
            # Abort fix if the masks differ
            warnings.warn(
                f"Extreme value(s) (potentially unmasked missing_values) found in {lon} and {lat} arrays: "
                f"{set(possible_missing_values)}. A fix is not possible since their locations are not consistent "
                "between the two arrays."
            )
            return fix

        # Check if there's only one unique extreme value
        if len(set(possible_missing_values)) == 1:
            fix = True
            missing_value = possible_missing_values[0]
            # Replace the missing value with np.NaN in place
            # and add _FillValue and missing_value attributes
            #  (ignoring already present attributes)
            for var in lat, lon:
                ds[var] = ds[var].where(ds[var] != missing_value, other=np.nan)
            if lat_bnds is not None and lon_bnds is not None:
                ds[lat_bnds] = ds[lat_bnds].where(
                    ds[lat] != missing_value, other=np.nan
                )
                ds[lon_bnds] = ds[lon_bnds].where(
                    ds[lon] != missing_value, other=np.nan
                )
            for var in [
                var for var in [lat, lon, lat_bnds, lon_bnds] if var is not None
            ]:
                ds[var].encoding["_FillValue"] = float(1e20)
                ds[var].encoding["missing_value"] = float(1e20)
                ds[var].attrs["_FillValue"] = float(1e20)
                ds[var].attrs["missing_value"] = float(1e20)
            warnings.warn(
                f"Unmasked missing_value found (and treated) in {lon} and {lat} arrays: '{missing_value}'."
            )
        else:
            # Raise warning - the values will likely cause an exception later on, depending on the operation
            warnings.warn(
                f"Multiple extreme values (potentially unmasked missing_values) found in {lon} and {lat} arrays: {set(possible_missing_values)}. This may cause issues."
            )

    return fix


def calculate_offset(lon, first_element_value):
    """Calculate the number of elements to roll the dataset by in order to have longitude from within requested bounds.

    Parameters
    ----------
    lon
        Longitude coordinate of xarray dataset.
    first_element_value
        The value of the first element of the longitude array to roll to.
    """
    # get resolution of data
    res = lon.values[1] - lon.values[0]

    # calculate how many degrees to move by to have lon[0] of rolled subset as lower bound of request
    diff = lon.values[0] - first_element_value

    # work out how many elements to roll by to roll data by 1 degree
    index = 1 / res

    # calculate the corresponding offset needed to change data by diff
    offset = int(round(diff * index))

    return offset


def _crosses_0_meridian(lon_c: xr.DataArray):
    """Determine whether grid extents over the 0-meridian.

    Assumes approximate constant width of grid cells.

    Parameters
    ----------
    lon_c: xr.DataArray
        Longitude coordinate variable in the longitude frame [-180, 180].

    Returns
    -------
    bool
        True for a dataset crossing the 0-meridian, False else.
    """
    if not isinstance(lon_c, xr.DataArray):
        raise InvalidParameterValue("Input needs to be of type xarray.DataArray.")

    # Not crossing the 0-meridian if all values are positive or negative
    lon_n = lon_c.where(lon_c <= 0, 0)
    lon_p = lon_c.where(lon_c >= 0, 0)
    if lon_n.all() or lon_p.all():
        return False

    # Determine min/max lon values
    xc_min = float(lon_c.min())
    xc_max = float(lon_c.max())

    # Determine resolution in zonal direction
    if lon_c.ndim == 1:
        xc_inc = (xc_max - xc_min) / (lon_c.sizes[lon_c.dims[0]] - 1)
    else:
        xc_inc = (xc_max - xc_min) / (lon_c.sizes[lon_c.dims[1]] - 1)

    # Generate a histogram with bins for sections along a latitudinal circle,
    #  width of the bins/sections dependent on the resolution in x-direction
    atol = 2.0 * xc_inc
    extent_hist = np.histogram(
        lon_c,
        bins=np.arange(xc_min - xc_inc, xc_max + atol, atol),
    )

    # If the counts for all bins are greater than zero, the grid is considered crossing the 0-meridian
    if np.all(extent_hist[0]):
        return True
    else:
        return False


def _convert_interval_between_lon_frames(low, high):
    """Convert a longitude interval to another longitude frame, returns Tuple of two floats."""
    diff = high - low
    if low < 0 and high > 0:
        raise ValueError(
            "Cannot convert longitude interval if it includes the 0°- or 180°-meridian."
        )
    elif low < 0:
        return tuple(sorted((low + 360.0, low + 360.0 + diff)))
    elif low < 180 and high > 180:
        raise ValueError(
            "Cannot convert longitude interval if it includes the 0°- or 180°-meridian."
        )
    elif high > 180:
        return tuple(sorted((high - 360.0 - diff, high - 360.0)))
    else:
        return float(low), float(high)


def cf_convert_between_lon_frames(ds_in, lon_interval, force=False):
    """Convert ds or lon_interval (whichever deems appropriate) to the other longitude frame, if the longitude frames do not match.

    If ds and lon_interval are defined on different longitude frames ([-180, 180] and [0, 360]),
    this function will convert one of the input parameters to the other longitude frame, preferably
    the lon_interval.
    Adjusts shifted longitude frames [0-x, 360-x] in the dataset to one of the two standard longitude
    frames, dependent on the specified lon_interval.
    In case of curvilinear grids featuring an additional 1D x-coordinate of the projection,
    this projection x-coordinate will not get converted.

    Parameters
    ----------
    ds_in: xarray.Dataset or xarray.DataArray
        xarray data object with defined longitude dimension.
    lon_interval: tuple or list
        length-2-tuple or -list of floats or integers denoting the bounds of the longitude interval.
    force: bool
        If True, force conversion even if longitude frames match.

    Returns
    -------
    Tuple(ds, lon_low, lon_high)
        The xarray.Dataset and the bounds of the longitude interval, potentially adjusted in terms
        of their defined longitude frame.
    """
    # Collect input specs
    if isinstance(ds_in, (xr.Dataset, xr.DataArray)):
        lon = detect_coordinate(ds_in, "longitude")
        lat = detect_coordinate(ds_in, "latitude")
        lon_bnds = detect_bounds(ds_in, lon)
        # lat_bnds = detect_bounds(ds_in, lat)
        # Do not consider bounds in gridtype detection (yet fails due to open_mfdataset bug that adds
        #  time dimension to bounds - todo)
        gridtype = detect_gridtype(
            ds_in, lon=lon, lat=lat
        )  # lat_bnds=lat_bnds, lon_bnds = lon_bnds)
        ds = ds_in.copy()
    else:
        raise InvalidParameterValue(
            "This function requires an xarray.DataArray or xarray.Dataset as input."
        )
    low, high = lon_interval
    lon_min, lon_max = ds.coords[lon].min().item(), ds.coords[lon].max().item()
    atol = 0.5

    # Conversion between longitude frames if required
    if (lon_min <= low or np.isclose(low, lon_min, atol=atol)) and (
        lon_max >= high or np.isclose(high, lon_max, atol=atol)
    ):
        if not force:
            return ds, low, high

    # Check longitude
    # For longitude frames other than [-180, 180] and [0, 360] in the dataset the following assumptions
    #  are being made:
    # - fixpoint is the 0-meridian
    # - the lon_interval is either defined in the longitude frame [-180, 180] or [0, 360]
    # TODO: possibly sth like
    #  while lon_min < -180, lon[lon<-180]=lon[lon<-180]+360
    #  while lon_max > 360, lon[lon>360]=lon[lon>360]-360
    if lon_max - lon_min > 360 + atol or lon_min < -360 - atol or lon_max > 360 + atol:
        raise ValueError(
            "The longitude coordinate values have to lie within the interval "
            "[-360, 360] degrees and not exceed an extent of 360 degrees."
        )

    # Conversion: longitude is a singleton dimension
    elif (ds[lon].ndim == 1 and ds.sizes[ds[lon].dims[0]] == 1) or (
        ds[lon].ndim > 1 and ds.sizes[ds[lon].dims[1]] == 1
    ):
        if low < 0 and lon_min > 0:
            ds[lon] = ds[lon].where(ds[lon] <= 180, ds[lon] - 360.0)
            if lon_bnds:
                ds[lon_bnds] = ds[lon_bnds].where(
                    ds[lon_bnds] <= 180, ds[lon_bnds] - 360.0
                )
        elif low > 0 and lon_min < 0:
            ds[lon] = ds[lon].where(ds[lon] >= 0, ds[lon] + 360.0)
            if lon_bnds:
                ds[lon_bnds] = ds[lon_bnds].where(
                    ds[lon_bnds] >= 0, ds[lon_bnds] + 360.0
                )
        return ds, low, high

    # Conversion: 1D or 2D longitude coordinate variable
    else:
        # regional [0 ... 180]
        if lon_min >= 0 and lon_max <= 180:
            return ds, low, high

        # shifted frame beyond -180, eg. [-300, 60]
        elif lon_min < -180 - atol:
            if low < 0:
                ds[lon] = ds[lon].where(ds[lon] > -180, ds[lon] + 360.0)
                if lon_bnds:
                    ds[lon_bnds] = ds[lon_bnds].where(
                        ds[lon_bnds] > -180, ds[lon_bnds] + 360.0
                    )
            elif low >= 0:
                ds[lon] = ds[lon].where(ds[lon] >= 0, ds[lon] + 360.0)
                if lon_bnds:
                    ds[lon_bnds] = ds[lon_bnds].where(
                        ds[lon_bnds] >= 0, ds[lon_bnds] + 360.0
                    )

        # shifted frame beyond 0, eg. [-60, 300]
        elif lon_min < -atol and lon_max > 180 + atol:
            if low < 0:
                ds[lon] = ds[lon].where(ds[lon] <= 180, ds[lon] - 360.0)
                if lon_bnds:
                    ds[lon_bnds] = ds[lon_bnds].where(
                        ds[lon_bnds] <= 180, ds[lon_bnds] - 360.0
                    )
            elif low >= 0:
                ds[lon] = ds[lon].where(ds[lon] >= 0, ds[lon] + 360.0)
                if lon_bnds:
                    ds[lon_bnds] = ds[lon_bnds].where(
                        ds[lon_bnds] >= 0, ds[lon_bnds] + 360.0
                    )

        # [-180 ... 180]
        elif lon_min < 0:
            # interval includes 180°-meridian: convert dataset to [0, 360]
            if low < 180 and high > 180:
                ds[lon] = ds[lon].where(ds[lon] >= 0, ds[lon] + 360.0)
                if lon_bnds:
                    ds[lon_bnds] = ds[lon_bnds].where(
                        ds[lon_bnds] >= 0, ds[lon_bnds] + 360.0
                    )
            # interval does not include 180°-meridian: convert interval to [-180,180]
            else:
                if low >= 0:
                    if not force:
                        low, high = _convert_interval_between_lon_frames(low, high)
                    else:
                        ds[lon] = ds[lon].where(ds[lon] >= 0, ds[lon] + 360.0)
                        if lon_bnds:
                            ds[lon_bnds] = ds[lon_bnds].where(
                                ds[lon_bnds] >= 0, ds[lon_bnds] + 360.0
                            )
                return ds, low, high

        # [0 ... 360]
        else:
            # interval positive, return unchanged
            if low >= 0:
                return ds, low, high
            # interval includes 0°-meridian: convert dataset to [-180, 180]
            elif high > 0:
                ds[lon] = ds[lon].where(ds[lon] <= 180, ds[lon] - 360.0)
                if lon_bnds:
                    ds[lon_bnds] = ds[lon_bnds].where(
                        ds[lon_bnds] <= 180, ds[lon_bnds] - 360.0
                    )
            # interval negative
            else:
                if not force:
                    low, high = _convert_interval_between_lon_frames(low, high)
                    return ds, low, high
                else:
                    ds[lon] = ds[lon].where(ds[lon] <= 180, ds[lon] - 360.0)
                    if lon_bnds:
                        ds[lon_bnds] = ds[lon_bnds].where(
                            ds[lon_bnds] <= 180, ds[lon_bnds] - 360.0
                        )
        # 1D coordinate variable: Sort, since order might no longer be ascending / descending
        if gridtype == "regular_lat_lon":
            ds = ds.sortby(lon)

        return ds, low, high


def check_lon_alignment(ds, lon_bnds):
    """Check whether the longitude subset requested is within the bounds of the dataset.

    If not try to roll the dataset so that the request is. Raise an exception if rolling is not possible.
    """
    low, high = lon_bnds
    lon = get_coord_by_type(ds, "longitude", ignore_aux_coords=False)
    lon = ds.coords[lon]
    lon_min, lon_max = lon.values.min(), lon.values.max()

    # handle the case where there is only one longitude
    if len(lon.values) == 1:
        lon_value = ds.lon.values[0]
        if low > lon_value:
            ds.coords[lon.name] = ds.coords[lon.name] + 360
        elif high < lon_value:
            ds.coords[lon.name] = ds.coords[lon.name] - 360
        return ds

    # check if the request is in bounds - return ds if it is
    if (lon_min <= low or np.isclose(low, lon_min, atol=0.5)) and (
        lon_max >= high or np.isclose(high, lon_max, atol=0.5)
    ):
        return ds

    else:
        # check if lon is a dimension
        if lon.name not in ds.dims:
            raise Exception(
                f"The requested longitude subset {lon_bnds} is not within the longitude bounds "
                f"of this dataset and the data could not be converted to this longitude frame successfully. "
                f"Please re-run your request with longitudes within the bounds of the dataset: ({lon_min:.2f}, {lon_max:.2f})"
            )
        # roll the dataset and reassign the longitude values
        else:
            first_element_value = low
            offset = calculate_offset(lon, first_element_value)

            # roll the dataset
            ds_roll = ds.roll(shifts={f"{lon.name}": offset}, roll_coords=True)

            # assign longitude to match the roll and copy attrs
            lon_vals = ds_roll.coords[lon.name].values

            # treat the values differently according to positive/negative offset
            if offset < 0:
                lon_vals[offset:] = lon_vals[offset:] % 360
            else:
                lon_vals[:offset] = lon_vals[:offset] % -360

            ds_roll.coords[lon.name] = lon_vals
            ds_roll.coords[lon.name].attrs = ds.coords[lon.name].attrs
            return ds_roll


def adjust_date_to_calendar(da, date, direction="backwards"):
    """Check that the date specified exists in the calendar type of the dataset.

    If not present, changes the date a day at a time (up to a maximum of 5 times) to find a date that does exist.
    The direction to change the date by is indicated by 'direction'.

    Parameters
    ----------
    da : xarray.Dataset or xarray.DataArray
        The data to examine.
    date : str
        The date to check.
    direction : str
        The direction to move in days to find a date that does exist.
        'backwards' means the search will go backwards in time until an existing date is found.
        'forwards' means the search will go forwards in time.
        The default is 'backwards'.

    Returns
    -------
    str
        The next possible existing date in the calendar of the dataset.
    """
    # turn date into AnyCalendarDateTime object
    d = str_to_AnyCalendarDateTime(date)

    # get the calendar type
    cal = da.cf["time"].data[0].calendar

    for i in range(5):
        try:
            cftime.datetime(
                d.year,
                d.month,
                d.day,
                d.hour,
                d.minute,
                d.second,
                calendar=cal,
            )
            return d.value
        except ValueError:
            if direction == "forwards":
                d.add_day()
            elif direction == "backwards":
                d.sub_day()
            else:
                raise Exception(
                    f"Invalid value for direction: {direction}. This should be either 'backwards' to indicate subtracting a day or 'forwards' for adding a day."
                )

    raise ValueError(
        f"Could not find an existing date near {date} in the calendar: {cal}"
    )


def add_hor_CF_coord_attrs(
    ds, lat="lat", lon="lon", lat_bnds="lat_bnds", lon_bnds="lon_bnds", keep_attrs=False
):
    """
    Add the common CF variable attributes to the horizontal coordinate variables.

    Parameters
    ----------
    lat : str, optional
        Latitude coordinate variable name. The default is "lat".
    lon : str, optional
        Longitude coordinate variable name. The default is "lon".
    lat_bnds : str, optional
        Latitude bounds coordinate variable name. The default is "lat_bnds".
    lon_bnds : str, optional
        Longitude bounds coordinate variable name. The default is "lon_bnds".
    keep_attrs : bool, optional
        Whether to keep original coordinate variable attributes if they do not conflict.
        In case of a conflict, the attribute value will be overwritten independent of this setting.
        The default is False.

    Returns
    -------
    xarray.Dataset
        The input dataset with updated coordinate variable attributes.
    """
    # Define common CF coordinate variable attrs
    lat_attrs = {
        "bounds": lat_bnds,
        "units": "degrees_north",
        "long_name": "latitude",
        "standard_name": "latitude",
        "axis": "Y",
    }
    lon_attrs = {
        "bounds": lon_bnds,
        "units": "degrees_east",
        "long_name": "longitude",
        "standard_name": "longitude",
        "axis": "X",
    }

    # Overwrite or update coordinate variables of input dataset
    try:
        if keep_attrs:
            ds[lat].attrs.update(lat_attrs)
            ds[lon].attrs.update(lon_attrs)
        else:
            ds[lat].attrs = lat_attrs
            ds[lon].attrs = lon_attrs
            ds[lat_bnds].attrs = {}
            ds[lon_bnds].attrs = {}
    except KeyError:
        raise KeyError("Not all specified coordinate variables exist in the dataset.")

    return ds


def reformat_SCRIP_to_CF(ds, keep_attrs=False):
    """Reformat dataset from SCRIP to CF format.

    Parameters
    ----------
    ds : xarray.Dataset
        Input dataset in SCRIP format.
    keep_attrs: bool
        Whether to keep the global attributes.

    Returns
    -------
    ds_ref : xarray.Dataset
        Reformatted dataset.
    """
    source_format = "SCRIP"
    target_format = "CF"
    SCRIP_vars = [
        "grid_center_lat",
        "grid_center_lon",
        "grid_corner_lat",
        "grid_corner_lon",
        "grid_dims",
        "grid_area",
        "grid_imask",
    ]

    if not isinstance(ds, xr.Dataset):
        raise InvalidParameterValue(
            "Reformat is only possible for Datasets."
            " DataArrays have to be CF conformal coordinate variables defined."
        )

    # Cannot reformat data variables yet
    if not (
        all([var in SCRIP_vars for var in ds.data_vars])
        and all([coord in SCRIP_vars for coord in ds.coords])
    ):
        raise Exception(
            "Converting the grid format from %s to %s is not yet possible for data variables."
            % (source_format, target_format)
        )

    # center lat and lon arrays will become the lat and lon arrays
    lat = ds.grid_center_lat.values.reshape(
        (ds.grid_dims.values[1], ds.grid_dims.values[0])
    ).astype(np.float32)
    lon = ds.grid_center_lon.values.reshape(
        (ds.grid_dims.values[1], ds.grid_dims.values[0])
    ).astype(np.float32)

    # corner coordinates will become lat_bnds and lon_bnds arrays
    # regular lat-lon case
    # todo: bounds of curvilinear case
    if all(
        [
            np.array_equal(lat[:, i], lat[:, i + 1], equal_nan=True)
            for i in range(ds.grid_dims.values[0] - 1)
        ]
    ) and all(
        [
            np.array_equal(lon[i, :], lon[i + 1, :], equal_nan=True)
            for i in range(ds.grid_dims.values[1] - 1)
        ]
    ):
        # regular lat-lon grid:
        # - 1D coordinate variables
        lat = lat[:, 0]
        lon = lon[0, :]
        # - reshape vertices from (n,2) to (n+1) for lat and lon axes
        lat_b = ds.grid_corner_lat.values.reshape(
            (
                ds.grid_dims.values[1],
                ds.grid_dims.values[0],
                ds.sizes["grid_corners"],
            )
        ).astype(np.float32)
        lon_b = ds.grid_corner_lon.values.reshape(
            (
                ds.grid_dims.values[1],
                ds.grid_dims.values[0],
                ds.sizes["grid_corners"],
            )
        ).astype(np.float32)
        lat_bnds = np.zeros((ds.grid_dims.values[1], 2), dtype=np.float32)
        lon_bnds = np.zeros((ds.grid_dims.values[0], 2), dtype=np.float32)
        lat_bnds[:, 0] = np.min(lat_b[:, 0, :], axis=1)
        lat_bnds[:, 1] = np.max(lat_b[:, 0, :], axis=1)
        lon_bnds[:, 0] = np.min(lon_b[0, :, :], axis=1)
        lon_bnds[:, 1] = np.max(lon_b[0, :, :], axis=1)
        ds_ref = xr.Dataset(
            data_vars={},
            coords={
                "lat": (["lat"], lat),
                "lon": (["lon"], lon),
                "lat_bnds": (["lat", "bnds"], lat_bnds),
                "lon_bnds": (["lon", "bnds"], lon_bnds),
            },
        )
        # todo: Case of other units (rad)
        # todo: Reformat data variables if in ds, apply imask on data variables
        # todo: vertical axis, time axis, ... ?!

        # add common coordinate variable attrs
        ds_ref = add_hor_CF_coord_attrs(ds=ds_ref)

        # transfer global attributes
        if keep_attrs:
            ds_ref.attrs.update(ds.attrs)

        return ds_ref
    else:
        raise Exception(
            "Converting the grid format from %s to %s is yet only possible for regular latitude longitude grids."
            % (source_format, target_format)
        )


def reformat_xESMF_to_CF(ds, keep_attrs=False):
    """Reformat dataset from xESMF to CF format.

    Parameters
    ----------
    ds : xarray.Dataset
        Input dataset in xESMF format.
    keep_attrs: bool
        Whether to keep the global attributes.

    Returns
    -------
    ds_ref : xarray.Dataset
        Reformatted dataset.
    """
    # source_format="xESMF"
    # target_format="CF"
    # todo: Check if it is regular_lat_lon, Check dimension sizes
    # Define lat, lon, lat_bnds, lon_bnds
    lat = ds.lat[:, 0]
    lon = ds.lon[0, :]
    lat_bnds = np.zeros((lat.shape[0], 2), dtype=np.float32)
    lon_bnds = np.zeros((lon.shape[0], 2), dtype=np.float32)

    # From (N+1, M+1) shaped bounds to (N, M, 4) shaped vertices
    lat_vertices = cfxr.vertices_to_bounds(ds.lat_b, ("bnds", "lat", "lon")).values
    lon_vertices = cfxr.vertices_to_bounds(ds.lon_b, ("bnds", "lat", "lon")).values

    # No longer necessary as of cf_xarray v0.7.5
    # lat_vertices = np.moveaxis(lat_vertices, 0, -1)
    # lon_vertices = np.moveaxis(lon_vertices, 0, -1)

    # From (N, M, 4) shaped vertices to (N, 2)  and (M, 2) shaped bounds
    lat_bnds[:, 0] = np.min(lat_vertices[:, 0, :], axis=1)
    lat_bnds[:, 1] = np.max(lat_vertices[:, 0, :], axis=1)
    lon_bnds[:, 0] = np.min(lon_vertices[0, :, :], axis=1)
    lon_bnds[:, 1] = np.max(lon_vertices[0, :, :], axis=1)

    # Create dataset
    ds_ref = xr.Dataset(
        data_vars={},
        coords={
            "lat": (["lat"], lat.data),
            "lon": (["lon"], lon.data),
            "lat_bnds": (["lat", "bnds"], lat_bnds.data),
            "lon_bnds": (["lon", "bnds"], lon_bnds.data),
        },
    )

    # todo: Case of other units (rad)
    # todo: Reformat data variables if in ds, apply imask on data variables
    # todo: vertical axis, time axis, ... ?!

    # add common coordinate variable attrs
    ds_ref = add_hor_CF_coord_attrs(ds=ds_ref)

    # transfer global attributes
    if keep_attrs:
        ds_ref.attrs.update(ds.attrs)

    return ds_ref
    #        else:
    #            raise Exception(
    #                "Converting the grid format from %s to %s is yet only possible for regular latitude longitude grids."
    #                % (self.format, grid_format)
    #            )


def detect_format(ds):
    """Detect format of a dataset. Yet supported are 'CF', 'SCRIP', 'xESMF'.

    Parameters
    ----------
    ds : xr.Dataset
        xarray.Dataset of which to detect the format.

    Returns
    -------
    str
        The format, if supported. Else raises an Exception.
    """
    # todo: extend for formats CF, xESMF, ESMF, UGRID, SCRIP
    # todo: add more conditions (dimension sizes, ...)
    SCRIP_vars = [
        "grid_center_lat",
        "grid_center_lon",
        "grid_corner_lat",
        "grid_corner_lon",
        # "grid_imask", "grid_area"
    ]
    SCRIP_dims = ["grid_corners", "grid_size", "grid_rank"]

    xESMF_vars = [
        "lat",
        "lon",
        "lat_b",
        "lon_b",
        # "mask",
    ]
    xESMF_dims = ["x", "y", "x_b", "y_b"]

    # Test if SCRIP
    if all([var in ds for var in SCRIP_vars]) and all(
        [dim in ds.dims for dim in SCRIP_dims]
    ):
        return "SCRIP"

    # Test if xESMF
    elif all([var in ds.coords for var in xESMF_vars]) and all(
        [dim in ds.dims for dim in xESMF_dims]
    ):
        return "xESMF"

    # Test if latitude and longitude can be found - standard_names would be set later if undef.
    elif (
        "latitude" in ds.cf.standard_names and "longitude" in ds.cf.standard_names
    ) or (
        get_coord_by_type(ds, "latitude", ignore_aux_coords=False) is not None
        and get_coord_by_type(ds, "longitude", ignore_aux_coords=False) is not None
    ):
        return "CF"

    else:
        raise Exception("The grid format is not supported.")


def detect_shape(ds, lat, lon, grid_type) -> tuple[int, int, int]:
    """Detect the shape of the grid.

    Returns a tuple of (nlat, nlon, ncells). For an unstructured grid nlat and nlon are not defined
    and therefore the returned tuple will be (ncells, ncells, ncells).

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing the grid / coordinate variables.
    lat : str
        Latitude variable name.
    lon : str
        Longitude variable name.
    grid_type: str
        One of "regular_lat_lon", "curvilinear", "unstructured"

    Returns
    -------
    int
        Number of latitude points in the grid.
    int
        Number of longitude points in the grid.
    int
        Number of cells in the grid.
    """
    if grid_type not in ["regular_lat_lon", "curvilinear", "unstructured"]:
        raise Exception(f"The specified grid_type '{grid_type}' is not supported.")

    if ds[lon].ndim != ds[lat].ndim:
        raise Exception(
            f"The coordinate variables {lat} and {lon} do not have the same number of dimensions."
        )
    elif ds[lat].ndim == 2:
        nlat = ds[lat].shape[0]
        nlon = ds[lat].shape[1]
        ncells = nlat * nlon
    elif ds[lat].ndim == 1:
        if ds[lat].shape == ds[lon].shape and grid_type == "unstructured":
            nlat = ds[lat].shape[0]
            nlon = nlat
            ncells = nlat
        else:
            nlat = ds[lat].shape[0]
            nlon = ds[lon].shape[0]
            ncells = nlat * nlon
    else:
        raise Exception(
            f"The coordinate variables {lat} and {lon} are not 1- or 2-dimensional."
        )
    return nlat, nlon, ncells


def _lonbnds_mids_trans_check(lon1, lon2, lon3, lon4):
    """Checks if the midpoints of the bounds traverse the Greenwich Meridian or
    antimeridian.If so, the midpoints are adjusted.
    """
    arr = np.array([lon1, lon2, lon3, lon4])
    diff = abs(arr.max() - arr.min())
    if diff > 180:
        # print("---------")
        # print(arr)
        arr = np.where(arr < 0.0, arr + 360.0, arr)
        # print(arr)
    mn = arr.mean()
    # print(mn)
    if mn > 180.0:
        mn = mn - 360.0
    # print(mn)
    return mn


def _lonbnds_mids_trans_check_diff(lon1, lon2):
    """Checks if the midpoints of the bounds traverse the Greenwich Meridian or
    antimeridian.If so, the midpoints are adjusted.
    """
    arr = np.array([lon1, lon2])
    if abs(arr[0] - arr[1]) > 180:
        arr = np.where(arr < 0.0, arr + 360.0, arr)
    val = arr[0] - (arr[1] - arr[0])
    if val > 180.0:
        val = val - 360.0
    return val


def _lonbnds_mids_trans_check_sum(lon1, lon2):
    """Checks if the midpoints of the bounds traverse the Greenwich Meridian or
    antimeridian.If so, the midpoints are adjusted.
    """
    arr = np.array([lon1, lon2])
    if abs(arr[0] - arr[1]) > 180:
        arr = np.where(arr < 0.0, arr + 360.0, arr)
    val = arr[0] + (arr[0] - arr[1])
    if val > 180.0:
        val = val - 360.0
    return val


def _determine_grid_orientation(lon):
    """Determine grid orientation by checking the longitude range along each axis."""
    # Compute the range of longitude values along both axes
    lon_range_axis_0 = abs(lon.max(axis=0) - lon.min(axis=0)).mean().item()
    lon_range_axis_1 = abs(lon.max(axis=1) - lon.min(axis=1)).mean().item()
    # print(lon_range_axis_0, lon_range_axis_1)

    if lon_range_axis_1 > lon_range_axis_0:
        return "nlat_nlon"  # Axis 1 corresponds to longitude (nlat, nlon)
    else:
        return "nlon_nlat"  # Axis 0 corresponds to longitude (nlon, nlat)


def generate_bounds_curvilinear(ds, lat, lon, clip_latitude=True, roll=True):
    """Compute bounds for curvilinear grids.

    Assumes 2D latitude and longitude coordinate variables. The bounds will be attached as coords
    to the xarray.Dataset. Assumes the longitudes are defined on the longitude frame [-180, 180].
    The default setting for 'roll' ensures that the longitudes
    are converted if this is not the case.

    The bound calculation for curvilinear grids was adapted from
    https://github.com/SantanderMetGroup/ATLAS/blob/mai-devel/scripts/ATLAS-data/\
    bash-interpolation-scripts/AtlasCDOremappeR_CORDEX/grid_bounds_calc.py
    which based on work by Caillaud Cécile and Samuel Somot from Meteo-France.
    Compared with the original code, there is an additional correction performed in the calculation,
    ensuring that at the Greenwich meridian and anti meridian the sign of the bounds does not differ.
    The new implementation is also significantly faster, as it replaces for loops with numpy.vectorize
    and index slicing.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to compute the bounds for.
    lon : str
        Longitude variable name.
    lat : str
        Latitude variable name.
    clip_latitude : bool, optional
        Whether to clip latitude values to [-90, 90]. The default is True.
    roll : bool, optional
        Whether to roll longitude values to [-180, 180]. The default is True.

    Returns
    -------
    ds : xarray.Dataset
        Dataset with bounds attached variables.
    """
    # Assume lon frame -180, 180
    if roll:
        ds, lonmin, lonmax = cf_convert_between_lon_frames(ds, (-180, 180), force=True)
    assert lonmin == -180 and lonmax == 180

    # Detect shape
    nlat, nlon, ncells = detect_shape(ds=ds, lat=lat, lon=lon, grid_type="curvilinear")

    lats = ds[lat].values
    lons = ds[lon].values

    orientation = _determine_grid_orientation(ds[lon])
    if orientation == "nlat_nlon":
        londim = 1
    else:
        nlon, nlat = nlat, nlon
        londim = 0
    # print(orientation)
    if londim == 0:
        lons_crnr = np.full((nlon + 1, nlat + 1), np.nan)
        lats_crnr = np.full((nlon + 1, nlat + 1), np.nan)
    else:
        lons_crnr = np.full((nlat + 1, nlon + 1), np.nan)
        lats_crnr = np.full((nlat + 1, nlon + 1), np.nan)

    if londim == 1 or londim == 0:
        lats_crnr[1:-1, 1:-1] = (
            lats[:-1, :-1] + lats[1:, :-1] + lats[:-1, 1:] + lats[1:, 1:]
        ) / 4.0
        lons_crnr[1:-1, 1:-1] = np.vectorize(
            lambda x1, x2, x3, x4: _lonbnds_mids_trans_check(x1, x2, x3, x4)
        )(lons[:-1, :-1], lons[1:, :-1], lons[:-1, 1:], lons[1:, 1:])

    # print(lons_crnr)

    # Grid points at boundaries - incl correction for cells crossing the prime/anti meridian
    lons_crnr[0, :] = np.vectorize(
        lambda x1, x2: _lonbnds_mids_trans_check_diff(x1, x2)
    )(lons_crnr[1, :], lons_crnr[2, :])
    # lons_crnr[1, :] - (lons_crnr[2, :] - lons_crnr[1, :])
    lons_crnr[-1, :] = np.vectorize(
        lambda x1, x2: _lonbnds_mids_trans_check_sum(x1, x2)
    )(lons_crnr[-2, :], lons_crnr[-3, :])
    # lons_crnr[-2, :] + (lons_crnr[-2, :] - lons_crnr[-3, :])
    lons_crnr[:, 0] = np.vectorize(
        lambda x1, x2: _lonbnds_mids_trans_check_diff(x1, x2)
    )(lons_crnr[:, 1], lons_crnr[:, 2])
    # lons_crnr[:, 1] - (lons_crnr[:, 2] - lons_crnr[:, 1])
    lons_crnr[:, -1] = np.vectorize(
        lambda x1, x2: _lonbnds_mids_trans_check_sum(x1, x2)
    )(lons_crnr[:, -2], lons_crnr[:, -3])
    # lons_crnr[:, -2] + (lons_crnr[:, -2] - lons_crnr[:, -3])

    lats_crnr[0, :] = lats_crnr[1, :] - (lats_crnr[2, :] - lats_crnr[1, :])
    lats_crnr[-1, :] = lats_crnr[-2, :] + (lats_crnr[-2, :] - lats_crnr[-3, :])
    lats_crnr[:, 0] = lats_crnr[:, 1] - (lats_crnr[:, 2] - lats_crnr[:, 1])
    lats_crnr[:, -1] = lats_crnr[:, -2] + (lats_crnr[:, -2] - lats_crnr[:, -3])

    if londim == 1:
        vertices_longitude = np.zeros((nlat, nlon, 4))
        vertices_latitude = np.zeros((nlat, nlon, 4))
    else:
        vertices_longitude = np.zeros((nlon, nlat, 4))
        vertices_latitude = np.zeros((nlon, nlat, 4))

    # Fill in counter clockwise
    vertices_longitude[:, :, 0] = lons_crnr[:-1, :-1]
    vertices_longitude[:, :, 1] = lons_crnr[:-1, 1:]
    vertices_longitude[:, :, 2] = lons_crnr[1:, 1:]
    vertices_longitude[:, :, 3] = lons_crnr[1:, :-1]
    vertices_latitude[:, :, 0] = lats_crnr[:-1, :-1]
    vertices_latitude[:, :, 1] = lats_crnr[:-1, 1:]
    vertices_latitude[:, :, 2] = lats_crnr[1:, 1:]
    vertices_latitude[:, :, 3] = lats_crnr[1:, :-1]

    # Clip latitudes
    if clip_latitude:
        vertices_latitude = np.clip(vertices_latitude, -90.0, 90.0)

    # Once more correct meridian crossing cells
    lon_range = vertices_longitude.max(axis=2) - vertices_longitude.min(axis=2)
    lon_range = np.repeat(lon_range[:, :, np.newaxis], 4, axis=2)

    # a=vertices_longitude[np.where(lon_range>180)]
    # b=vertices_longitude[np.where(vertices_longitude>180)]
    # print(a.shape)
    # print(a)
    # print("-----------------------------------")
    # print(b.shape)
    # print(b)
    # print("-----------------------------------")
    vertices_longitude = np.where(
        np.logical_and(lon_range > 180, vertices_longitude < 0),
        vertices_longitude + 360.0,
        vertices_longitude,
    )
    # lon_range = vertices_longitude.max(axis=2)-vertices_longitude.min(axis=2)
    # lon_range = np.repeat(lon_range[:, :, np.newaxis], 4, axis=2)
    # a=vertices_longitude[np.where(lon_range>180)]
    # print(a.shape)
    # print(a)
    # print("-----------------------------------")
    # print(b.shape)
    # print(b)

    # Add to the dataset
    ds["vertices_latitude"] = (
        (ds[lat].dims[0], ds[lat].dims[1], "vertices"),
        vertices_latitude,
    )
    ds["vertices_longitude"] = (
        (ds[lon].dims[0], ds[lon].dims[1], "vertices"),
        vertices_longitude,
    )
    ds[lat].attrs["bounds"] = "vertices_latitude"
    ds[lon].attrs["bounds"] = "vertices_longitude"

    return ds


def generate_bounds_rectilinear(ds, lat, lon):
    """Compute bounds for rectilinear grids.

    The bounds will be attached as coords to the xarray.Dataset of the Grid object.
    If no bounds can be created, a warning is issued. It is assumed but not ensured that no
    duplicated cells are present in the grid.

    Parameters
    ----------
    ds : xarray.Dataset
        .
    lat : str
        Latitude variable name.
    lon : str
        Longitude variable name.

    Returns
    -------
    xarray.Dataset
        Dataset with attached bounds.
    """
    # Detect shape
    nlat, nlon, ncells = detect_shape(
        ds=ds, lat=lat, lon=lon, grid_type="regular_lat_lon"
    )

    # Assuming lat / lon values are strong monotonically decreasing/increasing
    # Latitude / Longitude bounds shaped (nlat, 2) / (nlon, 2)
    lat_bnds = np.zeros((ds[lat].shape[0], 2), dtype=np.float32)
    lon_bnds = np.zeros((ds[lon].shape[0], 2), dtype=np.float32)

    # lat_bnds
    #  positive<0 for strong monotonically increasing
    #  positive>0 for strong monotonically decreasing
    positive = ds[lat].values[0] - ds[lat].values[1]
    gspacingl = abs(positive)
    gspacingu = abs(ds[lat].values[-1] - ds[lat].values[-2])
    if positive < 0:
        lat_bnds[1:, 0] = (ds[lat].values[:-1] + ds[lat].values[1:]) / 2.0
        lat_bnds[:-1, 1] = lat_bnds[1:, 0]
        lat_bnds[0, 0] = ds[lat].values[0] - gspacingl / 2.0
        lat_bnds[-1, 1] = ds[lat].values[-1] + gspacingu / 2.0
    elif positive > 0:
        lat_bnds[1:, 1] = (ds[lat].values[:-1] + ds[lat].values[1:]) / 2.0
        lat_bnds[:-1, 0] = lat_bnds[1:, 1]
        lat_bnds[0, 1] = ds[lat].values[0] + gspacingl / 2.0
        lat_bnds[-1, 0] = ds[lat].values[-1] - gspacingu / 2.0
    else:
        warnings.warn(
            "The bounds could not be calculated since the latitude and/or longitude "
            "values are not strong monotonically decreasing/increasing."
        )
        return ds

    lat_bnds = np.where(lat_bnds < -90.0, -90.0, lat_bnds)
    lat_bnds = np.where(lat_bnds > 90.0, 90.0, lat_bnds)

    # lon_bnds
    positive = ds[lon].values[0] - ds[lon].values[1]
    gspacingl = abs(positive)
    gspacingu = abs(ds[lon].values[-1] - ds[lon].values[-2])
    if positive < 0:
        lon_bnds[1:, 0] = (ds[lon].values[:-1] + ds[lon].values[1:]) / 2.0
        lon_bnds[:-1, 1] = lon_bnds[1:, 0]
        lon_bnds[0, 0] = ds[lon].values[0] - gspacingl / 2.0
        lon_bnds[-1, 1] = ds[lon].values[-1] + gspacingu / 2.0
    elif positive > 0:
        lon_bnds[1:, 1] = (ds[lon].values[:-1] + ds[lon].values[1:]) / 2.0
        lon_bnds[:-1, 0] = lon_bnds[1:, 1]
        lon_bnds[0, 1] = ds[lon].values[0] + gspacingl / 2.0
        lon_bnds[-1, 0] = ds[lon].values[-1] - gspacingu / 2.0
    else:
        warnings.warn(
            "The bounds could not be calculated since the latitude and/or longitude "
            "values are not strong monotonically decreasing/increasing."
        )
        return ds

    # Add to the dataset
    ds["lat_bnds"] = ((ds[lat].dims[0], "bnds"), lat_bnds)
    ds["lon_bnds"] = ((ds[lon].dims[0], "bnds"), lon_bnds)
    ds[lat].attrs["bounds"] = "lat_bnds"
    ds[lon].attrs["bounds"] = "lon_bnds"

    return ds


def detect_coordinate(ds, coord_type):
    """Use cf_xarray to obtain the variable name of the requested coordinate.

    Parameters
    ----------
    ds: xarray.Dataset, xarray.DataArray
        Dataset the coordinate variable name shall be obtained from.
    coord_type: str
        Coordinate type understood by cf-xarray, eg. 'lat', 'lon', ...

    Raises
    ------
    KeyError
        Raised if the requested coordinate cannot be identified.

    Returns
    -------
    str
        Coordinate variable name.
    """
    error_msg = f"A {coord_type} coordinate cannot be identified in the dataset."

    # Make use of cf-xarray accessor
    coord = get_coord_by_type(ds, coord_type, ignore_aux_coords=False)
    if coord is None:
        coord = get_coord_by_attr(ds, "standard_name", coord_type)
        if coord is None:
            raise KeyError(error_msg)

    # Return the name of the coordinate variable
    return coord


def detect_bounds(ds, coordinate) -> Optional[str]:
    """Use cf_xarray to obtain the variable name of the requested coordinates bounds.

    Parameters
    ----------
    ds : xarray.Dataset, xarray.DataArray
        Dataset the coordinate bounds variable name shall be obtained from.
    coordinate : str
        Name of the coordinate variable to determine the bounds from.

    Returns
    -------
    str or None
        Returns the variable name of the requested coordinate bounds.
        Returns None if the variable has no bounds or if they cannot be identified.
    """
    try:
        return ds.cf.bounds[coordinate][0]
    except (KeyError, AttributeError):
        warnings.warn(
            "For coordinate variable '%s' no bounds can be identified." % coordinate
        )
    return


def detect_gridtype(ds, lon, lat, lon_bnds=None, lat_bnds=None):
    """Detect type of the grid as one of "regular_lat_lon", "curvilinear", "unstructured".

    Assumes the grid description / structure follows the CF conventions.
    """
    # 1D coordinate variables
    if ds[lat].ndim == 1 and ds[lon].ndim == 1:
        lat_1D = ds[lat].dims[0]
        lon_1D = ds[lon].dims[0]
        if not lat_bnds or not lon_bnds:
            if lat_1D == lon_1D:
                return "unstructured"
            else:
                return "regular_lat_lon"
        else:
            # unstructured: bounds [ncells, nvertices]
            if (
                lat_1D == lon_1D
                and all([ds[bnds].ndim == 2 for bnds in [lon_bnds, lat_bnds]])
                and all(
                    [
                        ds.sizes[dim] > 2
                        for dim in [
                            ds[lon_bnds].dims[-1],
                            ds[lat_bnds].dims[-1],
                        ]
                    ]
                )
            ):
                return "unstructured"
            # rectilinear: bounds [nlat/nlon, 2]
            elif all([ds[bnds].ndim == 2 for bnds in [lon_bnds, lat_bnds]]) and all(
                [
                    ds.sizes[dim] == 2
                    for dim in [
                        ds[lon_bnds].dims[-1],
                        ds[lat_bnds].dims[-1],
                    ]
                ]
            ):
                return "regular_lat_lon"
            else:
                raise ValueError("The grid type is not supported.")

    # 2D coordinate variables
    elif ds[lat].ndim == 2 and ds[lon].ndim == 2:
        # Test for curvilinear or restructure lat/lon coordinate variables
        # todo: Check if regular_lat_lon despite 2D
        #  - requires additional function checking
        #      lat[:,i]==lat[:,j] for all i,j
        #      lon[i,:]==lon[j,:] for all i,j
        #  - and if that is the case to extract lat/lon and *_bnds
        #      lat[:]=lat[:,j], lon[:]=lon[j,:]
        #      lat_bnds[:, 2]=[min(lat_bnds[:,j, :]), max(lat_bnds[:,j, :])]
        #      lon_bnds similar
        if not ds[lat].shape == ds[lon].shape:
            raise ValueError(
                "The horizontal coordinate variables have differing shapes."
            )
        else:
            if not lat_bnds or not lon_bnds:
                return "curvilinear"
            else:
                # Shape of curvilinear bounds either [nlat, nlon, 4] or [nlat+1, nlon+1]
                if list(ds[lat].shape) + [4] == list(ds[lat_bnds].shape) and list(
                    ds[lon].shape
                ) + [4] == list(ds[lon_bnds].shape):
                    return "curvilinear"
                elif [si + 1 for si in ds[lat].shape] == list(ds[lat_bnds].shape) and [
                    si + 1 for si in ds[lon].shape
                ] == list(ds[lon_bnds].shape):
                    return "curvilinear"
                else:
                    raise ValueError("The grid type is not supported.")

    # >2D coordinate variables, or coordinate variables of different dimensionality
    else:
        raise ValueError(
            "The horizontal coordinate variables have more than 2 dimensions."
        )



import os
import sys
from pathlib import Path

import cf_xarray  # noqa
import pytest
import xarray as xr
from roocs_grids import get_grid_file, grid_dict

from clisops.core.regrid import XESMF_MINIMUM_VERSION, weights_cache_init, xe
from clisops.ops.regrid import regrid
from clisops.ops.subset import subset
from clisops.utils.testing import ContextLogger

XESMF_IMPORT_MSG = (
    f"xESMF >= {XESMF_MINIMUM_VERSION} is needed for regridding functionalities."
)


def _check_output_nc(result, fname="output_001.nc"):
    assert fname in [os.path.basename(_) for _ in result]


@pytest.mark.xfail(
    reason="Needs xarray fix https://github.com/pydata/xarray/issues/7794"
)
@pytest.mark.skipif(xe is None, reason=XESMF_IMPORT_MSG)
def test_regrid_basic(tmpdir, tmp_path, mini_esgf_data):
    """Test a basic regridding operation."""
    fpath = mini_esgf_data["CMIP5_MRSOS_ONE_TIME_STEP"]
    basename = os.path.splitext(os.path.basename(fpath))[0]
    method = "nearest_s2d"

    weights_cache_init(Path(tmp_path, "weights"))

    result = regrid(
        fpath,
        method=method,
        adaptive_masking_threshold=0.5,
        grid="1deg",
        output_dir=tmpdir,
        output_type="netcdf",
        file_namer="standard",
    )

    _check_output_nc(
        result, fname=f"{basename}-20051201_regrid-{method}-180x360_cells_grid.nc"
    )


@pytest.mark.skipif(xe is None, reason=XESMF_IMPORT_MSG)
def test_regrid_grid_as_none(tmpdir, tmp_path, mini_esgf_data):
    """Test behaviour when none passed as method and grid.

    Should use the default regridding.
    """
    fpath = mini_esgf_data["CMIP5_MRSOS_ONE_TIME_STEP"]

    weights_cache_init(Path(tmp_path, "weights"))

    with pytest.raises(
        Exception,
        match="xarray.Dataset, grid_id or grid_instructor have to be specified as input.",
    ):
        regrid(
            fpath,
            grid=None,
            output_dir=tmpdir,
            output_type="netcdf",
            file_namer="standard",
        )


@pytest.mark.slow
@pytest.mark.xfail(
    reason="Needs xarray fix https://github.com/pydata/xarray/issues/7794"
)
@pytest.mark.skipif(xe is None, reason=XESMF_IMPORT_MSG)
@pytest.mark.parametrize("grid_id", sorted(grid_dict))
def test_regrid_regular_grid_to_all_roocs_grids(
    tmpdir, tmp_path, grid_id, mini_esgf_data
):
    """Test for regridding a regular lat/lon field to all roocs grid types."""
    fpath = mini_esgf_data["CMIP5_MRSOS_ONE_TIME_STEP"]
    basename = os.path.splitext(os.path.basename(fpath))[0]
    method = "nearest_s2d"

    weights_cache_init(Path(tmp_path, "weights"))

    result = regrid(
        fpath,
        method=method,
        adaptive_masking_threshold=0.5,
        grid=grid_id,
        output_dir=tmpdir,
        output_type="netcdf",
        file_namer="standard",
    )

    nc_file = result[0]
    assert os.path.basename(nc_file).startswith(f"{basename}-20051201_regrid-{method}-")

    # Can we read the output file
    ds = xr.open_dataset(nc_file)
    assert "mrsos" in ds
    assert ds.mrsos.size > 100


@pytest.mark.slow
@pytest.mark.xfail(
    reason="Needs xarray fix https://github.com/pydata/xarray/issues/7794"
)
@pytest.mark.skipif(xe is None, reason=XESMF_IMPORT_MSG)
def test_subset_and_regrid_erroneous_cf_units_cmip5(tmpdir, mini_esgf_data, tmp_path):
    """Test subset and regrid ds with erroneous cf units."""
    fpath = mini_esgf_data["CMIP5_WRONG_CF_UNITS"]
    basename = os.path.splitext(os.path.basename(fpath))[0]
    method = "conservative"
    weights_cache_init(Path(tmp_path, "weights"))

    # subset
    result = subset(
        ds=fpath,
        area=(0, -10.0, 20.0, 10.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )
    _check_output_nc(result)

    # regrid
    result = regrid(
        result[0],
        method=method,
        adaptive_masking_threshold=0.5,
        grid="1deg",
        output_dir=tmpdir,
        output_type="netcdf",
        file_namer="standard",
    )

    nc_file = result[0]
    assert (
        os.path.basename(nc_file)
        == f"{basename.replace('Omon', 'mon')}16-20060116_regrid-{method}-180x360_cells_grid.nc"
    )

    # Can we read the output file
    ds = xr.open_dataset(nc_file)
    assert "zos" in ds
    assert ds.zos.size == 360 * 180
    assert ds.zos.count() == 163


@pytest.mark.slow
@pytest.mark.xfail(
    reason="Needs xarray fix https://github.com/pydata/xarray/issues/7794"
)
@pytest.mark.skipif(xe is None, reason=XESMF_IMPORT_MSG)
@pytest.mark.parametrize(
    "dset", ["ATLAS_v1_CORDEX", "ATLAS_v1_EOBS_GRID", "ATLAS_v0_CORDEX_ANT"]
)
def test_regrid_ATLAS_datasets(tmpdir, dset, mini_esgf_data):
    """Test regridding for several ATLAS datasets."""
    result = regrid(
        ds=mini_esgf_data[dset],
        method="bilinear",
        adaptive_masking_threshold=0.5,
        grid="0pt5deg_lsm",
        output_dir=tmpdir,
        output_type="netcdf",
        file_namer="standard",
    )
    assert os.path.basename(result[0]).endswith(
        "_regrid-bilinear-360x720_cells_grid.nc"
    )


@pytest.mark.xfail(
    reason="Needs xarray fix https://github.com/pydata/xarray/issues/7794"
)
@pytest.mark.skipif(xe is None, reason=XESMF_IMPORT_MSG)
def test_regrid_ATLAS_CORDEX(tmpdir, caplog, mini_esgf_data):  # noqa
    """Test regridding for ATLAS CORDEX dataset."""
    import netCDF4

    with ContextLogger(caplog) as _logger:
        _logger.add(sys.stdout, level="INFO")
        caplog.set_level("INFO", logger="clisops")

        _logger.info("netcdf4-python version: %s" % netCDF4.__version__)
        _logger.info("HDF5 lib version:       %s" % netCDF4.__hdf5libversion__)
        _logger.info("netcdf lib version:     %s" % netCDF4.__netcdf4libversion__)

    ds = xr.open_dataset(
        mini_esgf_data["ATLAS_v0_CORDEX_ANT"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    )

    # Might trigger KeyError in future netcdf-c versions
    # PR: https://github.com/Unidata/netcdf-c/pull/2716
    for cvar in [
        "member_id",
        "rcm_variant",
        "rcm_model",
        "rcm_institution",
        "gcm_variant",
        "gcm_model",
        "gcm_institution",
    ]:
        assert cvar in ds

    result = regrid(
        ds=ds,
        method="nearest_s2d",
        adaptive_masking_threshold=0.5,
        grid="1deg",
        output_dir=tmpdir,
        output_type="netcdf",
        file_namer="standard",
    )
    assert (
        os.path.basename(result[0])
        == "tnn_CORDEX-ANT_rcp45_mon_20060101-20060101_regrid-nearest_s2d-180x360_cells_grid.nc"
    )


@pytest.mark.skipif(xe is None, reason=XESMF_IMPORT_MSG)
def test_regrid_keep_attrs(tmp_path, mini_esgf_data):
    """Test if dataset and variable attributes are kept / removed as specified."""
    fpath = mini_esgf_data["CMIP6_TOS_ONE_TIME_STEP"]
    method = "nearest_s2d"

    weights_cache_init(Path(tmp_path, "weights"))

    ds = xr.open_dataset(fpath).isel(time=0)

    # regrid - preserve input attrs
    result = regrid(
        ds,
        method=method,
        adaptive_masking_threshold=-1,
        grid="2deg_lsm",
        output_type="xarray",
    )

    # regrid - scrapping attrs
    result_na = regrid(
        ds,
        method=method,
        adaptive_masking_threshold=-1,
        grid="2deg_lsm",
        output_type="xarray",
        keep_attrs=False,
    )

    # regrid - keep target attrs
    result_ta = regrid(
        ds,
        method=method,
        adaptive_masking_threshold=-1,
        grid="2deg_lsm",
        output_type="xarray",
        keep_attrs="target",
    )

    ds_remap = result[0]
    ds_remap_na = result_na[0]
    ds_remap_ta = result_ta[0]

    assert "tos" in ds_remap and "tos" in ds_remap_na and "tos" in ds_remap_ta
    assert all([key in ds_remap.tos.attrs.keys() for key in ds.tos.attrs.keys()])
    assert all(
        [
            key in ds_remap.attrs.keys()
            for key in ds.attrs.keys()
            if key not in ["nominal_resolution"]
        ]
    )
    # todo: remove the restriction when nominal_resolution of the target grid is calculated in core/regrid.py
    assert all([key not in ds_remap_na.tos.attrs.keys() for key in ds.tos.attrs.keys()])
    assert all(
        [
            key not in ds_remap_na.attrs.keys()
            for key in ds.attrs.keys()
            if key not in ["grid", "grid_label"]
        ]
    )
    assert all([key in ds_remap_ta.tos.attrs.keys() for key in ds.tos.attrs.keys()])
    assert all(
        [
            key not in ds_remap_ta.attrs.keys()
            for key in ds.attrs.keys()
            if key
            not in ["source", "Conventions", "history", "NCO", "grid", "grid_label"]
        ]
    )


@pytest.mark.slow
class TestRegridHalo:

    @pytest.mark.skipif(xe is None, reason=XESMF_IMPORT_MSG)
    def test_regrid_halo_simple(self, tmp_path, mini_esgf_data):
        """Test regridding with a simple halo."""
        fpath = mini_esgf_data["CMIP6_TOS_ONE_TIME_STEP"]
        ds = xr.open_dataset(fpath).isel(time=0)

        weights_cache_init(Path(tmp_path, "weights"))

        ds_out = regrid(
            ds,
            method="conservative",
            adaptive_masking_threshold=-1,
            grid=5,
            output_type="xarray",
        )[0]

        assert ds_out.attrs["regrid_operation"] == "conservative_404x802_36x72"

    @pytest.mark.skipif(xe is None, reason=XESMF_IMPORT_MSG)
    def test_regrid_halo_adv(self, tmp_path, mini_esgf_data):
        """Test regridding of dataset with a more complex halo."""
        fpath = mini_esgf_data["CMIP6_OCE_HALO_CNRM"]
        ds = xr.open_dataset(fpath).isel(time=0)

        weights_cache_init(Path(tmp_path, "weights"))

        ds_out = regrid(
            ds,
            method="conservative",
            adaptive_masking_threshold=-1,
            grid=5,
            output_type="xarray",
        )[0]

        assert ds_out.attrs["regrid_operation"] == "conservative_1050x1442_36x72"


@pytest.mark.skipif(xe is None, reason=XESMF_IMPORT_MSG)
def test_regrid_shifted_lon_frame(tmp_path, mini_esgf_data):
    """Test regridding of dataset with shifted longitude frame."""
    fpath = mini_esgf_data["CMIP6_IITM_EXTENT"]
    ds = xr.open_dataset(fpath).isel(time=0)

    weights_cache_init(Path(tmp_path, "weights"))

    ds_out = regrid(
        ds,
        method="bilinear",
        adaptive_masking_threshold=-1,
        grid=5,
        output_type="xarray",
    )[0]

    assert ds_out.attrs["regrid_operation"] == "bilinear_200x360_36x72_peri"


@pytest.mark.slow
@pytest.mark.skipif(xe is None, reason=XESMF_IMPORT_MSG)
def test_regrid_same_grid_exception(tmpdir, tmp_path):
    """Test that a warning is issued when source and target grid are the same."""
    fpath = get_grid_file("0pt25deg_era5")
    ds = xr.open_dataset(fpath)

    weights_cache_init(Path(tmp_path, "weights"))

    with pytest.warns(
        UserWarning,
        match="The selected source and target grids are the same.",
    ):
        ds_regrid = regrid(
            fpath,
            method="conservative",
            adaptive_masking_threshold=0.5,
            grid="0pt25deg_era5_lsm",
            output_dir=tmpdir,
            output_type="xarray",
            file_namer="standard",
        )[0]
    # It is expected that the input ds is simply passed through
    xr.testing.assert_identical(ds, ds_regrid)


@pytest.mark.xfail(
    reason="Needs xarray fix https://github.com/pydata/xarray/issues/7794"
)
@pytest.mark.skipif(xe is None, reason=XESMF_IMPORT_MSG)
def test_regrid_cmip6_nc_consistent_bounds_and_coords(tmpdir, mini_esgf_data):
    """Tests clisops regrid function and check metadata added by xarray"""
    result = regrid(
        ds=mini_esgf_data["CMIP6_ATM_VERT_ONE_TIMESTEP"],
        method="nearest_s2d",
        grid=10.0,
        output_dir=tmpdir,
        output_type="nc",
        file_namer="standard",
    )
    res = xr.open_mfdataset(result)
    # check fill value in bounds
    assert "_FillValue" not in res.lat_bnds.encoding
    assert "_FillValue" not in res.lon_bnds.encoding
    assert "_FillValue" not in res.time_bnds.encoding
    assert "_FillValue" not in res.lev_bnds.encoding
    assert "_FillValue" not in res.ap_bnds.encoding
    assert "_FillValue" not in res.b_bnds.encoding
    # check fill value in coordinates
    assert "_FillValue" not in res.time.encoding
    assert "_FillValue" not in res.lat.encoding
    assert "_FillValue" not in res.lon.encoding
    assert "_FillValue" not in res.lev.encoding
    assert "_FillValue" not in res.ap.encoding
    assert "_FillValue" not in res.b.encoding
    # check coordinates in bounds
    assert "coordinates" not in res.lat_bnds.encoding
    assert "coordinates" not in res.lon_bnds.encoding
    assert "coordinates" not in res.time_bnds.encoding
    assert "coordinates" not in res.lev_bnds.encoding
    assert "coordinates" not in res.ap_bnds.encoding
    assert "coordinates" not in res.b_bnds.encoding
    # Check coordinates not in variable attributes
    assert "coordinates" not in res.o3.encoding
    assert "coordinates" not in res.ps.encoding



"""CLISOPS - Climate simulation operations."""

######################################################################################
# BSD 3-Clause License
#
# Copyright (c) 2018-2025, United Kingdom Research and Innovation, Elle Smith, Ag Stephens, Carsten Ehbrecht, Trevor James Smith
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
######################################################################################

import warnings

from loguru import logger

from clisops.__version__ import __version__
from clisops.config import get_config
from clisops.utils.common import enable_logging


def showwarning(message, *args, **kwargs):
    """Inject warnings from `warnings.warn` into `loguru`."""
    logger.warning(message)
    showwarning_(message, *args, **kwargs)


showwarning_ = warnings.showwarning
warnings.showwarning = showwarning

# Disable logging for clisops and remove the logger that is instantiated on import
logger.disable("clisops")
logger.remove()

# Load configuration
CONFIG = get_config(__file__)

try:
    # Set the memory limit for each dask chunk
    chunk_memory_limit = CONFIG["clisops:read"].get("chunk_memory_limit", None)
except KeyError:
    logger.warning(
        "No chunk_memory_limit set in configuration file. Defaulting to None."
    )
    chunk_memory_limit = None

from clisops.parameter import *
from clisops.utils import *

for key, value in CONFIG["environment"].items():
    os.environ[key.upper()] = value



from __future__ import annotations

import os
from configparser import ConfigParser
from itertools import chain
from pathlib import Path

# Global _CONFIG used by other packages
_CONFIG = {}


def reload_config(package=None):
    """Reloads the configuration from the config file.

    Used for forcibly reloading the configuration from the config file, particularly useful for pytesting mock imports.
    """
    global _CONFIG  # noqa
    _load_config(package)

    for key, value in _CONFIG["environment"].items():
        os.environ[key.upper()] = value

    return _CONFIG


def get_config(package=None):
    global _CONFIG  # noqa

    if not _CONFIG:
        _load_config(package)

    return _CONFIG


def _gather_config_files(package: str | os.PathLike[str] | Path | None = None):
    conf_files = []
    _config = Path(__file__).parent.joinpath("etc").joinpath("roocs.ini")

    # add default config file
    # FIXME: we should be using importlib.resources to get the default config file
    if not _config.is_file():
        print(f"[WARN] Cannot load default config file from: {_config.as_posix()}")
    else:
        conf_files.append(_config)
    if package:
        pkg_config = Path(package).parent.joinpath("etc").joinpath("roocs.ini")
        if pkg_config.is_file():
            conf_files.append(pkg_config)

    # add system config /etc/roocs.ini
    sys_config = Path(Path(os.sep, "etc", "roocs.ini")).absolute()
    if sys_config.is_file():
        conf_files.append(sys_config)

    # add custom config from environment variable
    roocs_config = "ROOCS_CONFIG"
    if roocs_config in os.environ:
        conf_files.extend([Path(p) for p in os.environ[roocs_config].split(":")])

    return conf_files


def _to_list(i):
    return i.split()


def _to_dict(i):
    if not i.strip():
        return {}
    return dict([_.split(":") for _ in i.strip().split("\n")])


def _to_int(i):
    return int(i)


def _to_float(i):
    return float(i)


def _to_boolean(i):
    if i == "True":
        return True
    elif i == "False":
        return False
    else:
        raise ValueError(
            f"{i} is not valid for a boolean field - use 'True' or 'False'"
        )


def _chain_config_types(conf, keys):
    return chain(
        *[
            conf.get("config_data_types", key).split()
            for key in keys
            if conf.has_option("config_data_types", key)
        ]
    )


def _get_mappers(conf):
    mappers = {}

    for key in _chain_config_types(conf, ["lists", "extra_lists"]):
        mappers[key] = _to_list

    for key in _chain_config_types(conf, ["dicts", "extra_dicts"]):
        mappers[key] = _to_dict

    for key in _chain_config_types(conf, ["ints", "extra_ints"]):
        mappers[key] = _to_int

    for key in _chain_config_types(conf, ["floats", "extra_floats"]):
        mappers[key] = _to_float

    for key in _chain_config_types(conf, ["boolean", "extra_booleans"]):
        mappers[key] = _to_boolean

    return mappers


def _load_config(package=None):
    global _CONFIG

    conf_files = _gather_config_files(package)
    conf = ConfigParser()

    conf.read(conf_files)
    config = {}

    mappers = _get_mappers(conf)

    for section in conf.sections():
        config.setdefault(section, {})

        for key in conf.options(section):
            value = conf.get(section, key)

            if key in mappers:
                value = mappers[key](value)

            config[section][key] = value

    _post_process(config)

    _CONFIG = config


def _post_process(config):
    """
    Post-processes the contents of the config file to modify sections based on
    certain rules.
    Returns None.
    """
    for name in [n for n in config.keys() if n.startswith("project:")]:
        _modify_fixed_path_mappings(config, name)


def _modify_fixed_path_mappings(config, name):
    """
    Expands the contents of `fixed_path_mappings` based on other fixed path modifiers`.
    Returns None - changes contents in place.
    """
    d = config[name]

    fp_mappings = "fixed_path_mappings"
    fp_modifiers = "fixed_path_modifiers"

    if fp_mappings not in d or fp_modifiers not in d:
        return

    mappings = d[fp_mappings].copy()

    for modifier in d[fp_modifiers]:
        items = d[fp_modifiers][modifier].split()
        mappings = _expand_mappings(mappings, modifier, items)

    d[fp_mappings] = mappings.copy()


def _expand_mappings(mappings, modifier, items):
    """
    Expands mappings by replacing modifier with list of items
    in each case.
    """
    result = {}

    for key, value in mappings.items():
        lookup = "{" + modifier + "}"

        if lookup in key or lookup in value:
            for item in items:
                result[key.replace(lookup, item)] = value.replace(lookup, item)
        else:
            result[key] = value

    return result



from collections import ChainMap
from pathlib import Path
from typing import Optional, Union

import xarray as xr
from loguru import logger

from clisops.utils.common import expand_wildcards
from clisops.utils.dataset_utils import open_xr_dataset
from clisops.utils.file_namers import get_file_namer
from clisops.utils.output_utils import get_output, get_time_slices


class Operation:
    """Base class for all Operations."""

    def __init__(
        self,
        ds,
        file_namer: str = "standard",
        split_method: str = "time:auto",
        output_dir: Optional[Union[str, Path]] = None,
        output_type: str = "netcdf",
        **params,
    ):
        """Constructor for each operation.

        Sets common input parameters as attributes.
        Parameters that are specific to each operation are handled in `self._resolve_params()`
        """
        self._file_namer = file_namer
        self._split_method = split_method
        self._output_dir = output_dir
        self._output_type = output_type
        self._resolve_dsets(ds)
        self._resolve_params(**params)

    def _resolve_dsets(self, ds):
        """
        Take in the `ds` object and load it as an xarray Dataset if it
        is a path/wildcard. Set the result to `self.ds`.
        """
        if isinstance(ds, (str, Path)):
            ds = expand_wildcards(ds)
            ds = open_xr_dataset(ds)

        self.ds = ds

    def _resolve_params(self, **params) -> None:
        """
        Resolve the operation-specific input parameters to `self.params`.
        """
        self.params = params

    def _get_file_namer(self):
        """
        Return the appropriate file namer object.
        """
        namer = get_file_namer(self._file_namer)()
        return namer

    def _calculate(self):
        """The `_calculate()` method is implemented within each operation subclass."""
        raise NotImplementedError()

    def _remove_str_compression(self, ds):
        """
        netCDF4 datatypes of variable length are decoded to str by xarray<2023.11.0.
        As of xarray 2023.11.0 they are decoded to one of np.dtypes.StrDType (eg. "<U20")
        of variable length and stripped of all encoding settings. In netcdf-c versions >= 4.9.0
        and xarray < 2023.11.0 the latter part needs to be conducted manually to avoid an Exception
        when writing the xarray.Dataset to disk.
        See issue:  https://github.com/Unidata/netcdf4-python/issues/1205
        See PR: https://github.com/roocs/clisops/pull/319
        """
        if isinstance(ds, xr.Dataset):
            varlist = list(ds.coords) + list(ds.data_vars)
        elif isinstance(ds, xr.DataArray):
            varlist = list(ds.coords)

        for var in varlist:
            if "dtype" in ds[var].encoding:
                if ds[var].encoding["dtype"] == str:
                    for en in [
                        "compression",
                        "complevel",
                        "shuffle",
                        "fletcher32",
                        "endian",
                        "zlib",
                    ]:
                        if en in ds[var].encoding:
                            del ds[var].encoding[en]
        return ds

    def _cap_deflate_level(self, ds):
        """
        For CMOR3 / CMIP6 it was investigated which netCDF4 deflate_level should be set to optimize
        the balance between reduction of file size and degradation in performance. The values found
        were deflate_level=1, shuffle=True. To keep the write times at a minimum, compression level 1
        is not exceeded.
        See issue: https://github.com/PCMDI/cmor/issues/403
        """
        if isinstance(ds, xr.Dataset):
            varlist = list(ds.coords) + list(ds.data_vars)
        elif isinstance(ds, xr.DataArray):
            varlist = list(ds.coords)

        for var in varlist:
            complevel = ds[var].encoding.get("complevel", 0)
            compression = ds[var].encoding.get("compression_opts", 0)
            if complevel > 1:
                ds[var].encoding["complevel"] = 1
            elif compression > 1:
                ds[var].encoding["compression_opts"] = 1

        return ds

    @staticmethod
    def _remove_redundant_fill_values(ds):
        """Get coordinate and data variables and remove fill values added by xarray.

        CF-conventions say that coordinate variables cannot have missing values.

        See Also
        --------
        https://github.com/roocs/clisops/issues/224
        """
        if isinstance(ds, xr.Dataset):
            var_list = list(ds.coords) + list(ds.data_vars)
        elif isinstance(ds, xr.DataArray):
            var_list = list(ds.coords)
        else:
            raise ValueError(
                f"Expected xarray.Dataset or xarray.DataArray, got {type(ds)}"
            )

        for var in var_list:
            fval = ChainMap(ds[var].attrs, ds[var].encoding).get("_FillValue", None)
            mval = ChainMap(ds[var].attrs, ds[var].encoding).get("missing_value", None)
            if not fval and not mval:
                ds[var].encoding["_FillValue"] = None
            elif not mval:
                ds[var].encoding["missing_value"] = fval
                ds[var].encoding["_FillValue"] = fval
                ds[var].attrs.pop("_FillValue", None)
            elif not fval:
                ds[var].encoding["_FillValue"] = mval
                ds[var].encoding["missing_value"] = mval
                ds[var].attrs.pop("missing_value", None)
            else:
                # Issue 308 - Assert missing_value and _FillValue are the same
                if fval != mval:
                    ds[var].encoding["_FillValue"] = mval
                    ds[var].encoding["missing_value"] = mval
                    ds[var].attrs.pop("missing_value", None)
                    ds[var].attrs.pop("_FillValue", None)
                    logger.warning(
                        f"The defined _FillValue and missing_value for '{var}' are not the same '{fval}' != '{mval}'. Setting '{mval}' for both."
                    )
        return ds

    @staticmethod
    def _remove_redundant_coordinates_attr(ds):
        """This method removes the coordinates attribute added by xarray.

        Example
        -------
        .. code-block:: cpp

            double time_bnds(time, bnds);
            time_bnds:coordinates = "height";

        Programs like `cdo` will complain about this:

        .. code-block:: shell

            Warning (cdf_set_var): Inconsistent variable definition for time_bnds!

        See Also
        --------
        https://github.com/roocs/clisops/issues/224
        """
        if isinstance(ds, xr.Dataset):
            var_list = list(ds.coords) + list(ds.data_vars)
        elif isinstance(ds, xr.DataArray):
            var_list = list(ds.coords)
        else:
            raise ValueError(
                f"Expected xarray.Dataset or xarray.DataArray, got {type(ds)}"
            )

        for var in var_list:
            c_attr = ChainMap(ds[var].attrs, ds[var].encoding).get("coordinates", None)
            if not c_attr:
                ds[var].encoding["coordinates"] = None
            else:
                ds[var].encoding["coordinates"] = c_attr
                ds[var].attrs.pop("coordinates", None)
        return ds

    def process(self) -> list[Union[xr.Dataset, Path]]:
        """Main processing method used by all subclasses.

        Returns
        -------
        List[Union[xarray.Dataset, os.PathLike]]
            A list of outputs, which might be NetCDF file paths, Zarr file paths, or xarray.Dataset
        """
        # Create an empty list for outputs
        outputs = list()

        # Get the file namer object for naming output files
        # NOTE: It won't be used if the output type required is "xarray"
        namer = self._get_file_namer()

        # Process the xarray Dataset - this will (usually) be lazily evaluated so
        # no actual data will be read
        processed_ds = self._calculate()

        # remove fill values from lat/lon/time if required
        processed_ds = self._remove_redundant_fill_values(processed_ds)
        # remove redundant coordinates from bounds
        processed_ds = self._remove_redundant_coordinates_attr(processed_ds)
        # remove compression for string variables (as it is not supported by netcdf-c >= 4.9.0)
        processed_ds = self._remove_str_compression(processed_ds)
        # cap deflate level at 1
        processed_ds = self._cap_deflate_level(processed_ds)

        # Work out how many outputs should be created based on the size
        # of the array. Manage this as a list of time slices.
        time_slices = get_time_slices(processed_ds, self._split_method)

        # Loop through each time slice
        for tslice in time_slices:
            # If there is only one time slice, and it is None:
            # - then just set the result Dataset to the processed Dataset
            if tslice is None:
                result_ds = processed_ds
            # If there is a time slice then extract the time slice from the
            # processed Dataset
            else:
                result_ds = processed_ds.sel(time=slice(tslice[0], tslice[1]))

            logger.info(f"Processing {self.__class__.__name__} for times: {tslice}")

            # Get the output (file or xarray Dataset)
            # When this is a file: xarray will read all the data and write the file
            output = get_output(result_ds, self._output_type, self._output_dir, namer)
            outputs.append(output)

        return outputs



import pytest

from clisops.exceptions import InvalidParameterValue, MissingParameterValue
from clisops.parameter import CollectionParameter, collection

type_error = (
    "Input type of <{}> not allowed. Must be one of: "
    "[<class 'collections.abc.Sequence'>, <class 'str'>, <class 'clisops.parameter._utils.Series'>, <class 'clisops.utils.file_utils.FileMapper'>]"
)


def test__str__():
    coll = [
        "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
        "cmip5.output1.MPI-M.MPI-ESM-LR.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
    ]

    expected_str = (
        "Datasets to analyse:"
        "\ncmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga"
        "\ncmip5.output1.MPI-M.MPI-ESM-LR.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga"
    )

    parameter = CollectionParameter(coll)
    assert parameter.__str__() == expected_str
    assert parameter.__repr__() == parameter.__str__()
    assert parameter.__unicode__() == parameter.__str__()

    coll = collection(
        "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
        "cmip5.output1.MPI-M.MPI-ESM-LR.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
    )

    parameter = CollectionParameter(coll)
    assert parameter.__str__() == expected_str
    assert parameter.__repr__() == parameter.__str__()
    assert parameter.__unicode__() == parameter.__str__()


def test_raw():
    coll = [
        "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
        "cmip5.output1.MPI-M.MPI-ESM-LR.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
    ]
    parameter = CollectionParameter(coll)
    assert parameter.raw == coll


def test_validate_error_id():
    coll = [
        "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
        2,
    ]

    with pytest.raises(InvalidParameterValue) as exc:
        CollectionParameter(coll)
    assert (
        str(exc.value) == "Each id in a collection must be a string or "
        "an instance of <class 'clisops.utils.file_utils.FileMapper'>"
    )


def test_string():
    coll = (
        "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga,"
        "cmip5.output1.MPI-M.MPI-ESM-LR.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga"
    )

    parameter = CollectionParameter(coll)
    assert parameter.value == (
        "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
        "cmip5.output1.MPI-M.MPI-ESM-LR.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
    )


def test_one_id():
    coll = "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga"
    parameter = CollectionParameter(coll)
    assert parameter.value == (
        "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
    )


def test_whitespace():
    coll = (
        "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga, "
        "cmip5.output1.MPI-M.MPI-ESM-LR.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga "
    )

    parameter = CollectionParameter(coll)
    assert parameter.value == (
        "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
        "cmip5.output1.MPI-M.MPI-ESM-LR.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
    )


def test_empty_string():
    coll = ""
    with pytest.raises(MissingParameterValue) as exc:
        CollectionParameter(coll)
    assert str(exc.value) == "CollectionParameter must be provided"


def test_none():
    coll = None

    with pytest.raises(InvalidParameterValue) as exc:
        CollectionParameter(coll)
    assert str(exc.value) == type_error.format("class 'NoneType'")


def test_class_instance():
    coll = (
        "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga,"
        "cmip5.output1.MPI-M.MPI-ESM-LR.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga"
    )
    parameter = CollectionParameter(coll)
    new_parameter = CollectionParameter(parameter)
    assert new_parameter.value == (
        "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
        "cmip5.output1.MPI-M.MPI-ESM-LR.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga",
    )



import xarray as xr

from clisops.parameter import (
    area_parameter,
    collection_parameter,
    level_parameter,
    time_components_parameter,
    time_parameter,
)


def parameterise(
    collection=None, area=None, level=None, time=None, time_components=None
):
    """
    Parameterises inputs to instances of parameter classes which allows
    them to be used throughout roocs.
    For supported formats for each input please see their individual classes.

    :param collection: Collection input in any supported format.
    :param area: Area input in any supported format.
    :param level: Level input in any supported format.
    :param time: Time input in any supported format.
    :param time_components: Time Components input in any supported format.
    :return: Parameters as instances of their respective classes.
    """

    # if collection is a dataset/dataarray it doesn't need to be parameterised
    if type(collection) not in (xr.core.dataarray.DataArray, xr.core.dataset.Dataset):
        collection = collection_parameter.CollectionParameter(collection)

    area = area_parameter.AreaParameter(area)
    level = level_parameter.LevelParameter(level)
    time = time_parameter.TimeParameter(time)
    time_components = time_components_parameter.TimeComponentsParameter(time_components)

    return locals()



import os
import sys
import tempfile
import time
from pathlib import Path

import xarray as xr

from clisops import CONFIG
from clisops.utils.common import expand_wildcards
from clisops.utils.file_namers import get_file_namer
from clisops.utils.output_utils import (
    FileLock,
    get_chunk_length,
    get_da,
    get_output,
    get_time_slices,
)
from clisops.utils.testing import ContextLogger


def _open(coll):
    if isinstance(coll, (str, Path)):
        coll = expand_wildcards(coll)
    if len(coll) > 1:
        # issues with dask and cftime
        ds = xr.open_mfdataset(
            coll,
            decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
            combine="by_coords",
        ).load()
    else:
        ds = xr.open_dataset(
            coll[0], decode_times=xr.coders.CFDatetimeCoder(use_cftime=True)
        )
    return ds


def test_get_time_slices_single_slice(mini_esgf_data):
    tas = _open(mini_esgf_data["CMIP5_TAS"])

    test_data = [
        (
            tas,
            1000000,
            1,  # Setting file limit to 1000000 bytes
            ("2005-12-16", "2299-12-16"),
        ),
        (tas, None, 1, ("2005-12-16", "2299-12-16")),  # Using size limit from CONFIG
    ]

    split_method = "time:auto"
    for (
        ds,
        limit,
        n_times,
        slices,
    ) in test_data:
        resp = get_time_slices(ds, split_method, file_size_limit=limit)
        assert resp[0] == slices


def test_get_time_slices_multiple_slices(mini_esgf_data):
    tas = _open(mini_esgf_data["CMIP5_TAS"])

    test_data = [
        (
            tas,
            16000,
            4,
            ("2005-12-16", "2089-03-16"),
            ("2089-04-16", "2172-06-16"),
            ("2255-11-16", "2299-12-16"),
        ),
        (
            tas,
            32000,
            2,
            ("2005-12-16", "2172-06-16"),
            None,
            ("2172-07-16", "2299-12-16"),
        ),
    ]

    split_method = "time:auto"
    for ds, limit, n_times, first, second, last in test_data:
        resp = get_time_slices(ds, split_method, file_size_limit=limit)
        assert resp[0] == first
        assert resp[-1] == last

        if second:
            assert resp[1] == second


def test_tmp_dir_created_with_staging_dir(tmpdir):
    with ContextLogger() as _logger:
        _logger.add(sys.stdout, level="INFO")

        staging = Path(tmpdir).joinpath("tests")
        staging.mkdir(exist_ok=True)

        # copy part of function that creates tmp dir to check that it is created
        CONFIG["clisops:write"]["output_staging_dir"] = staging
        staging_dir = CONFIG["clisops:write"].get("output_staging_dir", "")

        output_path = "./output_001.nc"

        if os.path.isdir(staging_dir):
            tmp_dir = tempfile.TemporaryDirectory(dir=staging_dir)
            fname = os.path.basename(output_path)
            target_path = os.path.join(tmp_dir.name, fname)
            _logger.info(f"Writing to temporary path: {target_path}")
        else:
            target_path = output_path

        assert target_path != "output_001.nc"
        temp_test_folders = [f for f in staging.glob("tmp*")]
        assert len(temp_test_folders) == 1
        assert "tests/tmp" in temp_test_folders[0].as_posix()


def test_tmp_dir_not_created_with_no_staging_dir():
    # with ContextLogger() as _logger:
    #     _logger.add(sys.stdout, level="INFO")

    # copy part of function that creates tmp dir to check that it is not created when no staging dir
    CONFIG["clisops:write"]["output_staging_dir"] = ""
    staging_dir = CONFIG["clisops:write"].get("output_staging_dir", "")

    output_path = "./output_001.nc"

    if os.path.isdir(staging_dir):
        tmp_dir = tempfile.TemporaryDirectory(dir=staging_dir)
        fname = os.path.basename(output_path)
        target_path = os.path.join(tmp_dir.name, fname)
    else:
        target_path = output_path

    assert target_path == "./output_001.nc"


def test_no_staging_dir(caplog, mini_esgf_data):
    with ContextLogger(caplog) as _logger:
        _logger.add(sys.stdout, level="INFO")
        caplog.set_level("INFO", logger="clisops")

        CONFIG["clisops:write"]["output_staging_dir"] = ""
        ds = _open(mini_esgf_data["CMIP5_TAS"])
        output_path = get_output(
            ds, output_type="nc", output_dir=".", namer=get_file_namer("simple")()
        )

        assert "Writing to temporary path: " not in caplog.text
        assert output_path == "output_001.nc"

        os.remove("output_001.nc")


def test_invalid_staging_dir(caplog, mini_esgf_data):
    with ContextLogger(caplog) as _logger:
        _logger.add(sys.stdout, level="INFO")
        caplog.set_level("INFO", logger="clisops")

        # check staging dir not used with invalid directory
        CONFIG["clisops:write"]["output_staging_dir"] = "test/not/real/dir/"

        ds = _open(mini_esgf_data["CMIP5_TAS"])
        output_path = get_output(
            ds, output_type="nc", output_dir=".", namer=get_file_namer("simple")()
        )
        assert "Writing to temporary path: " not in caplog.text
        assert output_path == "output_001.nc"

        os.remove("output_001.nc")


def test_staging_dir_used(caplog, tmpdir, mini_esgf_data):
    with ContextLogger(caplog) as _logger:
        _logger.add(sys.stdout, level="INFO")
        caplog.set_level("INFO", logger="clisops")

        # check staging dir used when valid directory
        staging = Path(tmpdir).joinpath("tests")
        staging.mkdir(exist_ok=True)
        CONFIG["clisops:write"]["output_staging_dir"] = str(staging)
        ds = _open(mini_esgf_data["CMIP5_TAS"])

        output_path = get_output(
            ds, output_type="nc", output_dir=".", namer=get_file_namer("simple")()
        )

        assert f"Writing to temporary path: {staging}" in caplog.text
        assert output_path == "output_001.nc"

        Path("output_001.nc").unlink()


def test_final_output_path_staging_dir(mini_esgf_data):
    # check final output file in correct location with a staging directory used
    CONFIG["clisops:write"]["output_staging_dir"] = "tests/"

    ds = _open(mini_esgf_data["CMIP5_TAS"])
    get_output(ds, output_type="nc", output_dir=".", namer=get_file_namer("simple")())

    assert os.path.isfile("./output_001.nc")

    os.remove("output_001.nc")


def test_final_output_path_no_staging_dir(mini_esgf_data):
    # check final output file in correct location with a staging directory is not used
    ds = _open(mini_esgf_data["CMIP5_TAS"])
    get_output(ds, output_type="nc", output_dir=".", namer=get_file_namer("simple")())

    assert os.path.isfile("./output_001.nc")

    os.remove("output_001.nc")


def test_tmp_dir_deleted(tmpdir, mini_esgf_data):
    # check temporary directory under staging dir gets deleted after data has bee staged
    staging = Path(tmpdir).joinpath("tests")
    staging.mkdir(exist_ok=True)
    CONFIG["clisops:write"]["output_staging_dir"] = staging

    # CONFIG["clisops:write"]["output_staging_dir"] = "tests/"

    ds = _open(mini_esgf_data["CMIP5_TAS"])
    get_output(ds, output_type="nc", output_dir=".", namer=get_file_namer("simple")())

    # check that no tmpdir directories exist
    assert len([f for f in staging.glob("tmp*")]) == 0

    os.remove("output_001.nc")


def test_unify_chunks_cmip5(mini_esgf_data):
    """test unify chunks with a cmip5 example.

    da.unify_chunks() doesn't change da.chunks
    ds = ds.unify_chunks() doesn't appear to change ds.chunks in our case
    """
    # DataArray unify chunks method
    ds1 = _open(mini_esgf_data["CMIP5_TAS"])
    da = get_da(ds1)
    chunk_length = get_chunk_length(da)
    chunked_ds1 = ds1.chunk({"time": chunk_length})
    da.unify_chunks()

    # Dataset unify chunks method
    chunk_length = get_chunk_length(da)
    chunked_ds2 = ds1.chunk({"time": chunk_length})
    chunked_ds2_unified = chunked_ds2.unify_chunks()

    # test that da.unify_chunks hasn't changed ds.chunks
    assert chunked_ds1.chunks == chunked_ds2.chunks
    # test that ds = ds.unify_chunks hasn't changed ds.chunks
    assert chunked_ds2.chunks == chunked_ds2_unified.chunks


def test_unify_chunks_cmip6(mini_esgf_data):
    """Test unify chunks with a cmip6 example.

    da.unify_chunks() doesn't change da.chunks
    ds = ds.unify_chunks() doesn't appear to change ds.chunks in our case
    """
    # DataArray unify chunks method
    ds1 = _open(mini_esgf_data["CMIP6_TOS"])
    da = get_da(ds1)
    chunk_length = get_chunk_length(da)
    chunked_ds1 = ds1.chunk({"time": chunk_length})
    da.unify_chunks()

    # Dataset unify chunks method
    chunk_length = get_chunk_length(da)
    chunked_ds2 = ds1.chunk({"time": chunk_length})
    chunked_ds2_unified = chunked_ds2.unify_chunks()

    # test that da.unify_chunks hasn't changed ds.chunks
    assert chunked_ds1.chunks == chunked_ds2.chunks
    # test that ds = ds.unify_chunks hasn't changed ds.chunks
    assert chunked_ds2.chunks == chunked_ds2_unified.chunks


def test_filelock_simple(tmp_path):
    LOCK_FILE = Path(tmp_path, "test.lock")
    DATA_FILE = Path(tmp_path, "test.dat")

    lock = FileLock(LOCK_FILE)

    lock.acquire()
    try:
        time.sleep(1)
        assert os.path.isfile(LOCK_FILE)
        assert lock.state == "LOCKED"
        open(DATA_FILE, "a").write("1")
    finally:
        lock.release()

    time.sleep(1)
    assert not os.path.isfile(LOCK_FILE)


def test_filelock_already_locked(tmp_path):
    LOCK_FILE = Path(tmp_path, "test.lock")

    lock1 = FileLock(LOCK_FILE)
    lock2 = FileLock(LOCK_FILE)

    lock1.acquire(timeout=10)

    try:
        lock2.acquire(timeout=5)
    except Exception as exc:
        assert str(exc) == f"Could not obtain file lock on {LOCK_FILE}"



import sys

from clisops.utils.common import _logging_examples, enable_logging  # noqa
from clisops.utils.testing import ContextLogger


class TestLoggingFuncs:
    def test_logging_configuration(self, caplog):
        with ContextLogger(caplog):
            caplog.set_level("WARNING", logger="clisops")

            _logging_examples()  # noqa

            assert ("clisops.utils.common", 10, "1") not in caplog.record_tuples
            assert ("clisops.utils.common", 40, "4") in caplog.record_tuples

    def test_disabled_enabled_logging(self, capsys):
        with ContextLogger() as _logger:
            _logger.disable("clisops")

            # CLISOPS disabled
            _logger.add(sys.stderr, level="WARNING")
            _logger.add(sys.stdout, level="INFO")

            _logging_examples()  # noqa

            captured = capsys.readouterr()
            assert "WARNING" not in captured.err
            assert "INFO" not in captured.out

            # re-enable CLISOPS logging
            _logger.enable("clisops")

            _logging_examples()  # noqa

            captured = capsys.readouterr()
            assert "INFO" not in captured.err
            assert "WARNING" in captured.err
            assert "INFO" in captured.out

    def test_logging_enabler(self, capsys):
        with ContextLogger():
            _logging_examples()  # noqa

            captured = capsys.readouterr()
            assert "WARNING" not in captured.err
            assert "INFO" not in captured.out

            enable_logging()

            _logging_examples()  # noqa

            captured = capsys.readouterr()
            assert "INFO" not in captured.err
            assert "WARNING" in captured.err
            assert "INFO" in captured.out



import glob
import os

import xarray as xr
from loguru import logger

from clisops import CONFIG
from clisops.exceptions import InvalidProject
from clisops.utils.file_utils import FileMapper


class DatasetMapper:
    r"""
    Class to map to data path, dataset ID and files from any dataset input.

    Dset must be a string and can be input as:
        - A dataset ID: e.g. "cmip5.output1.INM.inmcm4.rcp45.mon.ocean.Omon.r1i1p1.latest.zostoga".
        - A file path: e.g. "/badc/cmip5/data/cmip5/output1/MOHC/HadGEM2-ES/rcp85/mon/atmos/Amon/r1i1p1/latest/tas/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc".
        - A path to a group of files: e.g. "/badc/cmip5/data/cmip5/output1/MOHC/HadGEM2-ES/rcp85/mon/atmos/Amon/r1i1p1/latest/tas/\*.nc".
        - A directory e.g. "/badc/cmip5/data/cmip5/output1/MOHC/HadGEM2-ES/rcp85/mon/atmos/Amon/r1i1p1/latest/tas".
        - An instance of the FileMapper class (that represents a set of files within a single directory).

    When force=True, if the project can not be identified, any attempt to use the base_dir of a project
    to resolve the data path will be ignored. Any of data_path, ds_id and files that can be set, will be set.
    """

    SUPPORTED_EXTENSIONS = (".nc", ".gz")

    def __init__(self, dset, project=None, force=False):
        self._project = project
        self.dset = dset

        self._base_dir = None
        self._ds_id = None
        self._data_path = None
        self._files = []

        self._parse(force)

    @staticmethod
    def _get_base_dirs_dict():
        projects = get_projects()
        base_dirs = {
            project: CONFIG[f"project:{project}"]["base_dir"] for project in projects
        }
        return base_dirs

    @staticmethod
    def _is_ds_id(dset):
        return dset.count(".") > 1

    def _deduce_project(self, dset):
        if isinstance(dset, str):
            if dset.startswith("/"):
                # by default this returns c3s-cmip6 not cmip6 (as they have the same base_dir)
                base_dirs_dict = self._get_base_dirs_dict()
                for project, base_dir in base_dirs_dict.items():
                    if (
                        dset.startswith(base_dir)
                        and CONFIG[f"project:{project}"].get("is_default_for_path")
                        is True
                    ):
                        return project

            elif self._is_ds_id(dset):
                return dset.split(".")[0].lower()

            # this will not return c3s project names
            elif dset.endswith(".nc") or os.path.isfile(dset):
                dset = xr.open_dataset(
                    dset, decode_times=xr.coders.CFDatetimeCoder(use_cftime=True)
                )
                return get_project_from_ds(dset)

        else:
            raise InvalidProject(
                f"The format of {dset} is not known and the project name could not be found."
            )

    def _parse(self, force):
        # if instance of FileMapper
        if isinstance(self.dset, FileMapper):
            dset = self.dset.dirpath
        else:
            dset = self.dset

        # set project and base_dir
        if not self._project:
            try:
                self._project = self._deduce_project(dset)
                self._base_dir = get_project_base_dir(self._project)
            except InvalidProject:
                logger.info("The project could not be identified")
                if not force:
                    raise InvalidProject(
                        "The project could not be identified and force was set to false"
                    )

        # get base_dir in the case where project has been supplied
        if not self._base_dir and self._project:
            self._base_dir = get_project_base_dir(self._project)

        # if a file, group of files or directory to files - find files
        if dset.startswith("/") or dset.endswith(".nc"):
            # if instance of FileMapper
            if isinstance(self.dset, FileMapper):
                self._files = self.dset.file_paths
                self._data_path = self.dset.dirpath

            if os.path.splitext(dset)[-1] in self.SUPPORTED_EXTENSIONS:
                if "*" in dset:
                    self._files = sorted(glob.glob(dset))
                else:
                    self._files.append(dset)

                # remove file extension to create data_path
                self._data_path = "/".join(dset.split("/")[:-1])

            # if base_dir identified, insert into data_path
            if self._base_dir:
                self._ds_id = ".".join(
                    self._data_path.replace(self._base_dir, self._project)
                    .strip("/")
                    .split("/")
                )

        # test if dataset id
        elif self._is_ds_id(dset):
            self._ds_id = dset

            mappings = CONFIG.get(f"project:{self.project}", {}).get(
                "fixed_path_mappings", {}
            )

            # If the dataset uses a fixed path mapping (from the config file) then use it
            if self._ds_id in mappings:
                data_path = mappings[self._ds_id]
                self._data_path = os.path.join(self._base_dir, data_path)

                # Use pattern of fixed file mapping as glob pattern
                self._files = sorted(glob.glob(self._data_path))

            # Default mapping is done by converting '.' characters to '/' separators in path
            else:
                self._data_path = os.path.join(
                    self._base_dir, "/".join(dset.split(".")[1:])
                )

        # use to data_path to find files if not set already
        if len(self._files) < 1:
            self._files = sorted(glob.glob(os.path.join(self._data_path, "*.nc")))
            print(self._data_path)
            print(self._files)

    @property
    def raw(self):
        """Raw dataset input."""
        return self.dset

    @property
    def data_path(self):
        """Dataset input converted to a data path."""
        return self._data_path

    @property
    def ds_id(self):
        """Dataset input converted to a ds id."""
        return self._ds_id

    @property
    def base_dir(self):
        """The base directory of the input dataset."""
        return self._base_dir

    @property
    def files(self):
        """The files found from the input dataset."""
        return self._files

    @property
    def project(self):
        """The project of the dataset input."""
        return self._project


def derive_dset(dset):
    """
    Derives the dataset path of the provided dset.

    :param dset: dset input of type described by DatasetMapper.
    :return: dataset path of input dataset.
    """
    return DatasetMapper(dset).data_path


def derive_ds_id(dset):
    """
    Derives the dataset id of the provided dset.

    :param dset: dset input of type described by DatasetMapper.
    :return: ds id of input dataset.
    """
    return DatasetMapper(dset).ds_id


def datapath_to_dsid(datapath):
    """
    Switches from dataset path to ds id.

    :param datapath: dataset path.
    :return: dataset id of input dataset path.
    """
    return DatasetMapper(datapath).ds_id


def dsid_to_datapath(dsid):
    """
    Switches from ds id to dataset path.

    :param dsid: dataset id.
    :return: dataset path of input dataset id.
    """
    return DatasetMapper(dsid).data_path


def dset_to_filepaths(dset, force=False):
    """
    Gets filepaths deduced from input dset.

    :param dset: dset input of type described by DatasetMapper.
    :param force: When True and if the project of the input dset cannot be identified, DatasetMapper will attempt to find the files anyway. Default is False.
    :return: File paths deduced from input dataset.
    """
    mapper = DatasetMapper(dset, force=force)
    return mapper.files


def switch_dset(dset):
    """
    Switches between dataset path and ds id.

    :param dset: either dataset path or dataset ID.
    :return: either dataset path or dataset ID - switched from the input.
    """
    if dset.startswith("/"):
        return datapath_to_dsid(dset)
    else:
        return dsid_to_datapath(dset)


def get_projects():
    """Gets all the projects available in the config."""
    return [_.split(":")[1] for _ in CONFIG.keys() if _.startswith("project:")]


def get_project_from_ds(ds):
    """
    Gets the project from an xarray Dataset/DataArray.

    :param ds: xarray Dataset/DataArray.
    :return: The project derived from the input dataset.
    """
    for project in get_projects():
        key = map_facet("project", project)
        if ds.attrs.get(key, "").lower() == project:
            return project


def get_project_name(dset):
    """
    Gets the project from an input dset.

    :param dset: dset input of type described by DatasetMapper.
    :return: The project derived from the input dataset.
    """
    if type(dset) in (xr.core.dataarray.DataArray, xr.core.dataset.Dataset):
        return get_project_from_ds(dset)  # will not return c3s dataset

    else:
        return DatasetMapper(dset).project


def map_facet(facet, project):
    """Return mapped facet value from config or facet name if not found."""
    # Return mapped value or the same facet name
    proj_mappings = CONFIG[f"project:{project}"]["mappings"]
    return proj_mappings.get(facet, facet)


def get_facet(facet_name, facets, project):
    """Get facet from project config"""
    return facets[map_facet(facet_name, project)]


def get_project_base_dir(project):
    """Get the base directory of a project from the config."""
    try:
        return CONFIG[f"project:{project}"]["base_dir"]
    except KeyError:
        raise InvalidProject("The project supplied is not known.")


def get_data_node_dirs_dict():
    """Get a dictionary of the data node roots used for retrieving original files."""
    projects = get_projects()
    data_node_dirs = {
        project: CONFIG[f"project:{project}"].get("data_node_root")
        for project in projects
        if CONFIG[f"project:{project}"].get("data_node_root")
    }
    return data_node_dirs


def get_project_from_data_node_root(url):
    """Identify the project from data node root by identifying the data node root in the input url."""
    data_node_dict = get_data_node_dirs_dict()
    project = None

    for proj, data_node_root in data_node_dict.items():
        if data_node_root in url:
            project = proj

    if not project:
        raise InvalidProject(
            f"The project could not be identified from the URL "
            f"{url} so it could not be mapped to a file path."
        )
    return project


def url_to_file_path(url):
    """Convert input url of an original file to a file path"""
    project = get_project_from_data_node_root(url)

    data_node_root = CONFIG.get(f"project:{project}", {}).get("data_node_root")
    base_dir = CONFIG.get(f"project:{project}", {}).get("base_dir")
    file_path = os.path.join(base_dir, url.partition(data_node_root)[2])

    return file_path
